#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.


from ..core import MeshNeuron, NeuronList
from .. import config, utils
import pandas as pd
import numpy as np

from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import lru_cache
from textwrap import dedent

err_msg = dedent("""
      Failed to import `caveclient` library. Please install using pip:

            pip install caveclient -U

      """)

try:
    from caveclient import CAVEclient
    import cloudvolume as cv
except ImportError:
    config.logger.error(err_msg)
    CAVEclient = None
    cv = None
except BaseException:
    raise


logger = config.get_logger(__name__)
dataset = None
SILENCE_FIND_MAT_VERSION = False


@lru_cache(None)
def get_cave_client(datastack="cortex65", server_address=None):
    """Get caveclient for given datastack.

    Parameters
    ----------
    datastack :     "cortex65" | "cortex35" | "layer 2/3" | "h01_c3_flat" | str
                    Which dataset to query. "cortex65", "cortex35" and "layer 2/3"
                    are internally mapped to the corresponding sources: for example,
                    "minnie65_public_vXXX" for "cortex65" where XXX is always the
                    most recent version).

    """
    if not CAVEclient:
        raise ImportError(err_msg)

    return CAVEclient(datastack, server_address)


@lru_cache(None)
def _get_cloudvol(url, cache=True):
    """Get (cached) CloudVolume for given segmentation.

    Parameters
    ----------
    url :     str

    """
    if not cv:
        raise ImportError(err_msg)

    return cv.CloudVolume(
        url, cache=cache, use_https=True, parallel=10, progress=False, fill_missing=True
    )

@lru_cache(None)
def _cloudvol_from_cv(client, **kwargs):
    """Get and cache CloudVolume from CAVEclient.

    Parameters
    ----------
    client :     CAVEclient
    **kwargs
                Additional arguments passed to `cloudvolume.CloudVolume`.

    """
    return client.info.segmentation_cloudvolume(**kwargs)


def _get_somas(root_ids, client, materialization="auto"):
    """Fetch somas based on nuclei segmentation for given neuron(s).

    Since this is a nucleus detection you will find that some neurons do
    not have an entry despite clearly having a soma. This is due to the
    "avocado problem" where the nucleus is separate from the rest of the
    soma.

    Important
    ---------
    This data currently only exists for the 'cortex65' datastack (i.e.
    "minnie65_public_vXXX").

    Parameters
    ----------
    root_ids  :     int | list of ints | None
                    Root ID(s) for which to fetch soma infos. Use
                    `None` to fetch complete list of annotated nuclei.
    table :         str
                    Which table to use for nucleus annotations. Also
                    see the `microns.list_annotation_tables()` function.
    datastack :     "cortex65" | "cortex35" | "layer 2/3" | "h01_c3_flat" | str
                    Which dataset to query. "cortex65", "cortex35" and "layer 2/3"
                    are internally mapped to the corresponding sources: for example,
                    "minnie65_public_vXXX" for "cortex65" where XXX is always the
                    most recent version).

    Returns
    -------
    DataFrame
                    Pandas DataFrame with nuclei (see Examples). Root IDs
                    without a nucleus will simply not have an entry in the
                    table.

    """
    # This property should be set by the dispatching function
    table = client.materialize.nucleus_table

    if materialization == "auto":
        if root_ids is not None:
            materialization = roots_to_mat(root_ids, client, verbose=False)
        else:
            # Use the most recent materialization
            materialization = None

    filter_in_dict = None
    if not isinstance(root_ids, type(None)):
        root_ids = utils.make_iterable(root_ids)
        filter_in_dict = {"pt_root_id": root_ids}

    return client.materialize.query_table(
        table, filter_in_dict=filter_in_dict, materialization_version=materialization
    )


def fetch_neurons(
    x,
    lod,
    with_synapses,
    client,
    parallel,
    max_threads,
    materialization="auto",
    **kwargs,
):
    """Fetch neuron meshes.

    Notes
    -----
    Synapses will be attached to the closest vertex on the mesh.

    Parameters
    ----------
    x :             str | int | list-like
                    Segment ID(s). Multiple Ids can be provided as list-like.
    lod :           int
                    Level of detail. Higher ``lod`` = coarser. This parameter
                    is ignored if the data source does not support multi-level
                    meshes.
    with_synapses : bool, optional
                    If True will also attach synapses as ``.connectors``.
    client :        CAVEclient
                    The CAVEclient with which to interact.
    parallel :      bool
                    If True, will use parallel threads to fetch data.
    max_threads :   int
                    Max number of parallel threads to use.
    materialization : "auto" | int
                    Which materialization version to use to look up somas and synapses
                    (if applicable). If "auto" (default) will try to find the most
                    recent version that contains the given root IDs. If an
                    integer is provided will use that version.
    **kwargs
                    Keyword arguments are passed through to the initialization
                    of the ``navis.MeshNeurons``.

    Returns
    -------
    navis.Neuronlist
                    Containing :class:`navis.MeshNeuron`.

    """
    x = utils.make_iterable(x, force_type=int)

    # Let CAVEclient handle the cloudvolume (this should take care of any secrets)
    vol = _cloudvol_from_cv(client)

    try:
        somas = _get_somas(x, client=client, materialization=materialization)
        soma_pos = somas.set_index("pt_root_id").pt_position.to_dict()
    except BaseException as e:
        logger.warning(f"Failed to fetch somas via nucleus segmentation(){e})")
        soma_pos = {}

    nl = []
    if max_threads > 1 and parallel:
        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            futures = {}
            for id in x:
                f = executor.submit(
                    _fetch_single_neuron,
                    id,
                    vol=vol,
                    lod=lod,
                    client=client,
                    with_synapses=with_synapses,
                    materialization=materialization,
                    **kwargs,
                )
                futures[f] = id

            with config.tqdm(
                desc="Fetching",
                total=len(x),
                leave=config.pbar_leave,
                disable=len(x) == 1 or config.pbar_hide,
            ) as pbar:
                for f in as_completed(futures):
                    id = futures[f]
                    pbar.update(1)
                    try:
                        nl.append(f.result())
                    except Exception as exc:
                        print(f"{id} generated an exception:", exc)
    else:
        for id in config.tqdm(
            x,
            desc="Fetching",
            leave=config.pbar_leave,
            disable=len(x) == 1 or config.pbar_hide,
        ):
            n = _fetch_single_neuron(
                id,
                vol=vol,
                lod=lod,
                client=client,
                with_synapses=with_synapses,
                materialization=materialization,
                **kwargs,
            )
            nl.append(n)

    nl = NeuronList(nl)

    for n in nl:
        if n.id in soma_pos:
            # For VoxelResolution see client.materialize.get_table_metadata('nucleus_detection_v0')
            # (attached to df as 'table_voxel_resolution')
            n.soma_pos = (
                np.array(soma_pos[n.id]) * somas.attrs["table_voxel_resolution"]
            )
        else:
            n.soma_pos = None

    return nl


def _fetch_single_neuron(
    id, lod, vol, client, with_synapses=False, materialization="auto", **kwargs
):
    """Fetch a single neuron."""
    # Make sure we only use `lod` if that's actually supported by the source
    if "MultiLevel" in str(type(vol.mesh)):
        mesh = vol.mesh.get(id, lod=lod)[id]
    else:
        mesh = vol.mesh.get(id, deduplicate_chunk_boundaries=False)[id]

    n = MeshNeuron(mesh, id=id, units="nm", **kwargs)

    if materialization == "auto":
        materialization = roots_to_mat(id, client, verbose=False)

    if with_synapses:
        pre = client.materialize.synapse_query(
            pre_ids=id, materialization_version=materialization
        )
        post = client.materialize.synapse_query(
            post_ids=id, materialization_version=materialization
        )

        syn_table = client.materialize.synapse_table
        syn_info = client.materialize.get_table_metadata(syn_table)
        vxl_size = np.array(syn_info["voxel_resolution"]).astype(int)

        to_concat = []
        if not pre.empty:
            pre["type"] = "pre"
            locs = np.vstack(pre["pre_pt_position"].values)
            locs = locs * vxl_size
            pre["x"], pre["y"], pre["z"] = locs[:, 0], locs[:, 1], locs[:, 2]
            pre = pre[["id", "x", "y", "z", "type", "size"]].copy()
            to_concat.append(pre)
        if not post.empty:
            post["type"] = "post"
            locs = np.vstack(post["post_pt_position"].values)
            locs = locs * vxl_size
            post["x"], post["y"], post["z"] = locs[:, 0], locs[:, 1], locs[:, 2]
            post = post[["id", "x", "y", "z", "type", "size"]].copy()
            to_concat.append(post)

        if len(to_concat) == 1:
            n.connectors = to_concat[0]
        elif len(to_concat) == 2:
            n.connectors = pd.concat(to_concat, axis=0).reset_index(drop=True)

    return n


def get_voxels(x, mip, bounds, client):
    """Fetch voxels making a up given root ID.

    Parameters
    ----------
    x :             int
                    A single root ID.
    mip :           int
                    Scale at which to fetch voxels.
    bounds :        list, optional
                    Bounding box [xmin, xmax, ymin, ymax, zmin, zmax] in voxel
                    space. For example, the voxel resolution for mip 0
                    segmentation is 8 x 8 x 40 nm.
    client :        CAVEclient
                    The CAVEclient with which to interact.

    Returns
    -------
    voxels :        (N, 3) np.ndarray
                    In voxel space according to `mip`.

    """
    # Need to get the graphene (not the precomputed) version of the data
    vol_graphene = cv.CloudVolume(
        client.chunkedgraph.cloudvolume_path, use_https=True, progress=False
    )
    url = client.info.get_datastack_info()["segmentation_source"]
    vol_prec = _get_cloudvol(url)

    # Get L2 chunks making up this neuron
    l2_ids = client.chunkedgraph.get_leaves(x, stop_layer=2)

    # Turn l2_ids into chunk indices
    l2_ix = [
        np.array(vol_graphene.mesh.meta.meta.decode_chunk_position(l2)) for l2 in l2_ids
    ]
    l2_ix = np.unique(l2_ix, axis=0)

    # Convert to nm
    l2_nm = np.asarray(_chunks_to_nm(l2_ix, vol=vol_graphene))

    # Convert back to voxel space (according to mip)
    l2_vxl = l2_nm // vol_prec.meta.scales[mip]["resolution"]

    voxels = []
    ch_size = np.array(vol_graphene.mesh.meta.meta.graph_chunk_size)
    ch_size = ch_size // (vol_prec.mip_resolution(mip) / vol_prec.mip_resolution(0))
    ch_size = np.asarray(ch_size).astype(int)
    old_mip = vol_prec.mip

    if not isinstance(bounds, type(None)):
        bounds = np.asarray(bounds)
        if not bounds.ndim == 1 or len(bounds) != 6:
            raise ValueError("`bounds` must be [xmin, xmax, ymin, ymax, zmin, zmax]")
        l2_vxl = l2_vxl[np.all(l2_vxl >= bounds[::2], axis=1)]
        l2_vxl = l2_vxl[np.all(l2_vxl < bounds[1::2] + ch_size, axis=1)]

    try:
        vol_prec.mip = mip
        for ch in config.tqdm(l2_vxl, desc="Loading"):
            ct = vol_prec[
                ch[0] : ch[0] + ch_size[0],
                ch[1] : ch[1] + ch_size[1],
                ch[2] : ch[2] + ch_size[2],
            ][:, :, :, 0]
            this_vxl = np.dstack(np.where(ct == x))[0]
            this_vxl = this_vxl + ch
            voxels.append(this_vxl)
    except BaseException:
        raise
    finally:
        vol_prec.mip = old_mip
    voxels = np.vstack(voxels)

    if not isinstance(bounds, type(None)):
        voxels = voxels[np.all(voxels >= bounds[::2], axis=1)]
        voxels = voxels[np.all(voxels < bounds[1::2], axis=1)]

    return voxels


def _chunks_to_nm(xyz_ch, vol, voxel_resolution=[4, 4, 40]):
    """Map a chunk location to Euclidean space.

    Parameters
    ----------
    xyz_ch :            array-like
                        (N, 3) array of chunk indices.
    vol :               cloudvolume.CloudVolume
                        CloudVolume object associated with the chunked space.
    voxel_resolution :  list, optional
                        Voxel resolution.

    Returns
    -------
    np.array
                        (N, 3) array of spatial points.

    """
    mip_scaling = vol.mip_resolution(0) // np.array(voxel_resolution, dtype=int)

    x_vox = np.atleast_2d(xyz_ch) * vol.mesh.meta.meta.graph_chunk_size
    return (
        (x_vox + np.array(vol.mesh.meta.meta.voxel_offset(0)))
        * voxel_resolution
        * mip_scaling
    )


def roots_to_mat(
    ids,
    client,
    verbose=True,
    allow_multiple=False,
    raise_missing=True,
):
    """Find a materialization version (or live) for given root ID(s).

    Parameters
    ----------
    ids :           int | iterable
                    Root ID(s) to check.
    client :        CAVEclient
                    The CAVEclient with which to interact.
    verbose :       bool
                    Whether to print results of search.
    allow_multiple : bool
                    If True, will track if IDs can be found spread across multiple
                    materialization versions if there is no single one containing
                    all.
    raise_missing : bool
                    Only relevant if `allow_multiple=True`. If False, will return
                    versions even if some IDs could not be found.

    Returns
    -------
    version :       int | "live"
                    A single version (including "live") that contains all given
                    root IDs.
    versions :      np.ndarray
                    If no single version was found and `allow_multiple=True` will
                    return a vector of `len(ids)` with the latest version at which
                    the respective ID can be found.
                    Important: "live" version will be return as -1!
                    If `raise_missing=False` and one or more root IDs could not
                    be found in any of the available materialization versions
                    these IDs will be return as version 0.

    """
    ids = utils.make_iterable(ids)

    # For each ID track the most recent valid version
    latest_valid = np.zeros(len(ids), dtype=np.int32)

    # Get the meta data for the available materialization versions
    # This is a list of dicts where each dict has a "time_stamp" key
    vmeta = client.materialize.get_versions_metadata()

    # Sort by "time_stamp"
    vmeta = sorted(vmeta, key=lambda x: x["time_stamp"], reverse=True)

    # Go over each version (start with the most recent)
    for i, mat in enumerate(vmeta):
        ts_m = mat["time_stamp"]
        version = mat["version"]

        # Check which root IDs were valid at the time
        is_valid = client.chunkedgraph.is_latest_roots(ids, timestamp=ts_m)

        # Update latest valid versions
        latest_valid[(latest_valid == 0) & is_valid] = version

        if all(is_valid):
            if verbose and not SILENCE_FIND_MAT_VERSION:
                print(f"Using materialization version {version}.")
            return version

    # If no single materialized version can be found, see if we can get
    # by with the live materialization
    is_latest = client.chunkedgraph.is_latest_roots(ids, timestamp=None)
    latest_valid[(latest_valid == 0) & is_latest] = -1  # track "live" as -1
    if all(is_latest) and dataset != "public":  # public does not have live
        if verbose:
            print("Using live materialization")
        return "live"

    if allow_multiple and any(latest_valid != 0):
        if all(latest_valid != 0):
            if verbose and not SILENCE_FIND_MAT_VERSION:
                print(
                    f"Found root IDs spread across {len(np.unique(latest_valid))} "
                    "materialization versions."
                )
            return latest_valid

        msg = (
            f"Found root IDs spread across {len(np.unique(latest_valid)) - 1} "
            f"materialization versions but {(latest_valid == 0).sum()} IDs "
            "do not exist in any of the materialized tables."
        )

        if not raise_missing:
            if verbose and not SILENCE_FIND_MAT_VERSION:
                print(msg)
            return latest_valid
        else:
            raise MaterializationMatchError(msg)

    if dataset not in ("public, "):
        raise MaterializationMatchError(
            "Given root IDs do not (co-)exist in any of the available "
            "materialization versions (including live). Try updating "
            "root IDs and rerun your query."
        )
    else:
        raise MaterializationMatchError(
            "Given root IDs do not (co-)exist in any of the available "
            "public materialization versions. Please make sure that "
            "the root IDs do exist and rerun your query."
        )


class MaterializationMatchError(Exception):
    pass



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

"""Module contains functions to plot neurons in 3D."""

import os
import warnings

import numpy as np

from typing import Union, List
from importlib.util import find_spec

from .. import utils, config, core
from .colors import prepare_colormap
from .settings import OctarineSettings, PlotlySettings, K3dSettings, VispySettings

__all__ = ["plot3d"]

logger = config.get_logger(__name__)

# Check if backends are available without importing them
BACKENDS = tuple(
    b for b in ("octarine", "vispy", "plotly", "k3d") if find_spec(b) is not None
)
JUPYTER_BACKENDS = tuple(b for b in ("plotly", "octarine", "k3d") if b in BACKENDS)
NON_JUPYTER_BACKENDS = tuple(
    b for b in ("octarine", "vispy", "plotly") if b in BACKENDS
)
AUTO_BACKEND = None  # choose the backend only the first time


def plot3d(
    x: Union[
        core.NeuronObject,
        core.Volume,
        np.ndarray,
        List[Union[core.NeuronObject, np.ndarray, core.Volume]],
    ],
    **kwargs,
):
    """Generate interactive 3D plot.

    Uses either [octarine], [vispy], [k3d] or [plotly] as backend.
    By default, the choice is automatic depending on what backends
    are installed and the context:

      - Terminal: octarine > vispy > plotly
      - Jupyter: plotly > octarine > k3d

    See the `backend` parameter on how to change this behavior.

    [octarine]: https://schlegelp.github.io/octarine/
    [vispy]: http://vispy.org
    [k3d]: https://k3d-jupyter.org/
    [plotly]: http://plot.ly

    Parameters
    ----------
    x :               Neuron/List | Volume | numpy.array | list thereof
                      The object(s) to plot. Can be:
                        - navis neurons, neuronlists or volumes
                        - numpy.array (N,3) is plotted as scatter plot
                        - multiple objects can be passed as list (see examples)
                      See parameters below for ways to customize the plot.

    Object parameters
    -----------------
    color :           None | str | tuple | list | dict, default=None

                      Use single str (e.g. `'red'`) or `(r, g, b)` tuple
                      to give all neurons the same color. Use `list` of
                      colors to assign colors: `['red', (1, 0, 1), ...].
                      Use `dict` to map colors to neurons:
                      `{neuron.id: (r, g, b), ...}`.

    palette :         str | array | list of arrays, default=None

                      Name of a matplotlib or seaborn palette. If `color` is
                      not specified will pick colors from this palette.

    alpha :           float [0-1], optional

                      Alpha value for neurons. Overriden if alpha is provided
                      as color specified in `color` has an alpha channel.

    connectors :      bool | "presynapses" | "postsynapses" | str | list, default=True

                      Plot connectors. This can either be `True` (plot all
                      connectors), `"presynapses"` (only presynaptic connectors)
                      or `"postsynapses"` (only postsynaptic connectors). If
                      a string or a list is provided, it will be used to filter the
                      `type` column in the connectors table.

                      Use these parameters to adjust the way connectors are plotted:

                        - `cn_colors` (str | tuple | dict | "neuron" ) overrides
                          the default connector (e.g. synpase) colors:
                            - single color as str (e.g. `'red'`) or rgb tuple
                              (e.g. `(1, 0, 0)`)
                            - dict mapping the connectors tables `type` column to
                              a color (e.g. `{"pre": (1, 0, 0)}`)
                            - with "neuron", connectors will receive the same color
                              as their neuron
                        - `cn_layout` (dict): Layout of the connectors. See
                          `navis.config.default_connector_colors` for options.
                        - `cn_size` (float): Size of the connectors.
                        - `cn_alpha` (float): Transparency of the connectors.
                        - `cn_mesh_colors` (bool): Whether to color the connectors
                          by the neuron's color.

    connectors_only : bool, default=False

                      Plot only connectors (e.g. synapses) if available and
                      ignore the neurons.

    color_by :        str | array | list of arrays, default = None

                      Color neurons by a property. Can be:

                        - a list/array of labels, one per each neuron
                        - a neuron property (str)
                        - a column name in the node table of `TreeNeurons`
                        - a list/array of values for each node

                      Numerical values will be normalized. You can control
                      the normalization by passing a `vmin` and/or `vmax`
                      parameter. Must specify a colormap via `palette`.

    shade_by :        str | array | list of arrays, default=None

                      Similar to `color_by` but will affect only the alpha
                      channel of the color. If `shade_by='strahler'` will
                      compute Strahler order if not already part of the node
                      table (TreeNeurons only). Numerical values will be
                      normalized. You can control the normalization by passing
                      a `smin` and/or `smax` parameter. Does not work with
                      `k3d` backend.

    radius :          bool | "auto", default=False

                      If "auto" will plot neurites of `TreeNeurons` with radius
                      if they have radii. If True, will try plotting neurites of
                      `TreeNeurons` with radius regardless. The radius can be
                      scaled by `linewidth`. Note that this will increase rendering
                      time.

    soma :            bool, default=True

                      TreeNeurons only: Whether to plot soma if it exists. Size
                      of the soma is determined by the neuron's `.soma_radius`
                      property which defaults to the "radius" column for
                      `TreeNeurons`.

    linewidth :       float, default=3 for plotly and 1 for all others

                      TreeNeurons only.

    linestyle :       str, default='-'

                      TreeNeurons only. Follows the same rules as in matplotlib.

    scatter_kws :     dict, optional

                      Use to modify scatter plots. Accepted parameters are:
                        - `size` to adjust size of dots
                        - `color` to adjust color

    Figure parameters
    -----------------
    backend :         'auto' (default) | 'octarine' | 'vispy' | 'plotly' | 'k3d'

                      Which backend to use for plotting. Note that there will
                      be minor differences in what feature/parameters are
                      supported depending on the backend:

                        - `auto` selects backend based on availability and
                          context (see above). You can override this by setting an
                          environment variable e.g. `NAVIS_PLOT3D_BACKEND="vispy"`
                          or `NAVIS_PLOT3D_JUPYTER_BACKEND="k3d"`.
                        - `octarine` uses WGPU to generate high performances
                          interactive 3D plots. Works both terminal and Jupyter.
                        - `vispy` similar to octarine but uses OpenGL: slower
                          but runs on older systems. Works only from terminals.
                        - `plotly` generates 3D plots using WebGL. Works
                          "inline" in Jupyter notebooks but can also produce a
                          HTML file that can be opened in any browers.
                        - `k3d` generates 3D plots using k3d. Works only in
                          Jupyter notebooks!

    **Below parameters are for plotly backend only:**

    fig :             plotly.graph_objs.Figure

                      Pass to add graph objects to existing plotly figure. Will
                      not change layout.

    title :           str, default=None

                      For plotly only! Change plot title.

    width/height :    int, optional

                      Use to adjust figure size.

    fig_autosize :    bool, default=False

                      For plotly only! Autoscale figure size.
                      Attention: autoscale overrides width and height

    hover_name :      bool, default=False

                      If True, hovering over neurons will show their label.

    hover_id :        bool, default=False

                      If True, hovering over skeleton nodes will show their ID.

    legend :          bool, default=True

                      Whether or not to show the legend.

    legend_orientation : "v" (default) | "h"

                      Orientation of the legend. Can be 'h' (horizontal) or 'v'
                      (vertical).

    legend_group :    dict, default=None

                      A dictionary mapping neuron IDs to labels (strings).
                      Use this to group neurons under a common label in the
                      legend.

    inline :          bool, default=True

                      If True and you are in an Jupyter environment, will
                      render plotly/k3d plots inline. If False, will generate
                      and return either a plotly Figure or a k3d Plot object
                      without immediately showing it.

    **Below parameters are for the Octarine/vispy backends only:**

    clear :           bool, default = False

                      If True, will clear the viewer before adding the new
                      objects.

    center :          bool, default = True

                      If True, will center camera on the newly added objects.

    combine :         bool, default = False

                      If True, will combine objects of the same type into a
                      single visual. This can greatly improve performance but
                      also means objects can't be selected individually
                      anymore. This is Vispy only.

    size :            (width, height) tuple, optional

                      Use to adjust figure/window size.

    show :            bool, default=True

                      Whether to immediately show the viewer.

    Returns
    -------
    If `backend='octarine'`

        From terminal: opens a 3D window and returns :class:`octarine.Viewer`.
        From Jupyter: :class:`octarine.Viewer` displayed in an ipywidget.

    If `backend='vispy'`

        Opens a 3D window and returns [`navis.Viewer`][].

    If `backend='plotly'`

        Returns either `None` if you are in a Jupyter notebook (see also
        `inline` parameter) or a `plotly.graph_objects.Figure`
        (see examples).

    If `backend='k3d'`

        Returns either `None` and immediately displays the plot or a
        `k3d.plot` object that you can manipulate further (see `inline`
        parameter).

    See Also
    --------
    [`octarine.Viewer`](https://schlegelp.github.io/octarine/)
        Interactive 3D viewer.

    [`navis.Viewer`][]
        Interactive vispy 3D viewer.

    Examples
    --------
    >>> import navis

    In a Jupyter notebook using plotly as backend:

    >>> nl = navis.example_neurons()
    >>> # Backend is automatically chosen but we can set it explicitly
    >>> # Plot inline
    >>> nl.plot3d(backend='plotly')                             # doctest: +SKIP
    >>> # Plot as separate html in a new window
    >>> fig = nl.plot3d(backend='plotly', inline=False)
    >>> import plotly.offline
    >>> _ = plotly.offline.plot(fig)                            # doctest: +SKIP

    In a Jupyter notebook using k3d as backend:

    >>> nl = navis.example_neurons()
    >>> # Plot inline
    >>> nl.plot3d(backend='k3d')                                # doctest: +SKIP

    In a terminal using octarine as backend:

    >>> # Plot list of neurons
    >>> nl = navis.example_neurons()
    >>> v = navis.plot3d(nl, backend='octarine')                # doctest: +SKIP
    >>> # Clear canvas
    >>> navis.clear3d()

    Some more advanced examples:

    >>> # plot3d() can deal with combinations of objects
    >>> nl = navis.example_neurons()
    >>> vol = navis.example_volume('LH')
    >>> vol.color = (255, 0, 0, .5)
    >>> # This plots a neuronlists, a single neuron and a volume
    >>> v = navis.plot3d([nl[0:2], nl[3], vol])
    >>> # Clear viewer (works only with octarine and vispy)
    >>> v = navis.plot3d(nl, clear=True)

    See the [plotting intro](../../generated/gallery/1_plotting/tutorial_plotting_00_intro)
    for even more examples.

    """
    # Select backend
    backend = kwargs.pop("backend", "auto")
    allowed_backends = ("auto", "octarine", "vispy", "plotly", "k3d")
    if backend.lower() == "auto":
        global AUTO_BACKEND
        if AUTO_BACKEND is not None:
            backend = AUTO_BACKEND
        else:
            if utils.is_jupyter():
                if not len(JUPYTER_BACKENDS):
                    raise ModuleNotFoundError(
                        "No 3D plotting backends available for Jupyter "
                        "environment. Please install one of the following: "
                        "plotly, octarine, k3d."
                    )
                backend = os.environ.get(
                    "NAVIS_PLOT3D_JUPYTER_BACKEND", JUPYTER_BACKENDS[0]
                )
            else:
                if not len(NON_JUPYTER_BACKENDS):
                    raise ModuleNotFoundError(
                        "No 3D plotting backends available for REPL/script. Please "
                        "install one of the following: octarine, vispy, plotly."
                    )
                backend = os.environ.get(
                    "NAVIS_PLOT3D_BACKEND", NON_JUPYTER_BACKENDS[0]
                )

            # Set the backend for the next time
            AUTO_BACKEND = backend

            logger.info(f'Using "{backend}" backend for 3D plotting.')
    elif backend.lower() not in allowed_backends:
        raise ValueError(
            f'Unknown backend "{backend}". ' f'Permitted: {".".join(allowed_backends)}.'
        )
    elif backend.lower() not in BACKENDS:
        raise ModuleNotFoundError(
            f'Backend "{backend}" not installed. Please install it via pip '
            "(see https://navis.readthedocs.io/en/latest/source/install.html#optional-dependencies "
            "for more information)."
        )

    if backend == "vispy":
        return plot3d_vispy(x, **kwargs)
    elif backend == "k3d":
        if not utils.is_jupyter():
            logger.warning("k3d backend only works in Jupyter environments")
        return plot3d_k3d(x, **kwargs)
    elif backend == "plotly":
        return plot3d_plotly(x, **kwargs)
    elif backend == "octarine":
        return plot3d_octarine(x, **kwargs)
    else:
        raise ValueError(
            f'Unknown backend "{backend}". ' f'Permitted: {".".join(allowed_backends)}.'
        )


def plot3d_vispy(x, **kwargs):
    """Plot3d() helper function to generate vispy 3D plots.

    This is just to improve readability. Its only purpose is to find the
    existing viewer or generate a new one.

    """
    from .vispy.viewer import Viewer

    # If this likely the first invoke, warn the user that vispy is deprecated
    if not hasattr(config, "primary_viewer"):
        warnings.warn(
            (
                "The `vispy` backend is depcrecated and will be removed in a future version of navis. "
                "We recommend to use the `octarine` backend instead. If that is for some reason not possible, "
                "please let us know via the issue tracker at https://github.com/navis-org/navis/issues asap."
            ),
            DeprecationWarning,
            stacklevel=2,
        )

    settings = VispySettings().update_settings(**kwargs)

    # Parse objects to plot
    (neurons, volumes, points, visuals) = utils.parse_objects(x)

    if settings.viewer in (None, "new"):
        # If does not exists yet, initialise a canvas object and make global
        if (
            not isinstance(getattr(config, "primary_viewer", None), Viewer)
            or settings.viewer == "new"
        ):
            viewer = config.primary_viewer = Viewer(size=settings.size)
        else:
            viewer = getattr(config, "primary_viewer", None)
    else:
        viewer = settings.viewer

    # Make sure viewer is visible
    if settings.show:
        viewer.show()

    # We need to pop clear/clear3d to prevent clearing again later
    if settings.clear:
        settings.clear = False
        viewer.clear()

    # Add objects (the viewer currently takes care of producing the visuals)
    if neurons:
        viewer.add(neurons, **settings.to_dict())
    if volumes:
        viewer.add(volumes, **settings.to_dict())
    if points:
        viewer.add(points, scatter_kws=settings.scatter_kws)

    return viewer


def plot3d_octarine(x, **kwargs):
    """Plot3d() helper function to generate octarine 3D plots.

    This is just to improve readability. Its only purpose is to find the
    existing viewer or generate a new one.

    """
    # Lazy import because octarine is not a hard dependency
    try:
        import octarine as oc
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "navis.plot3d() with the `octarine` backend requires the `octarine3d` library "
            "to be installed:\n  pip3 install octarine3 octarine-navis-plugin -U"
        )

    if not hasattr(oc.Viewer, "add_neurons"):
        raise ModuleNotFoundError(
            "Looks like the navis plugin for octarine is not installed. "
            "Please install it via pip:\n  pip install octarine-navis-plugin"
        )

    settings = OctarineSettings().update_settings(**kwargs)

    # Parse objects to plot
    (neurons, volumes, points, visuals) = utils.parse_objects(x)

    # Check if any existing viewer has already been closed
    if isinstance(getattr(config, "primary_viewer", None), oc.Viewer):
        try:
            getattr(config, "primary_viewer").canvas.__repr__()
        except RuntimeError:
            config.primary_viewer = None

    if settings.viewer in (None, "new"):
        # If it does not exists yet, initialize a canvas object and make global
        if (
            not isinstance(getattr(config, "primary_viewer", None), oc.Viewer)
            or settings.viewer == "new"
        ):
            viewer = config.primary_viewer = oc.Viewer(
                size=settings.size,
                camera=settings.camera,
                control=settings.control,
                show=False,
                offscreen=settings.offscreen or os.environ.get("NAVIS_HEADLESS", False),
            )
        else:
            viewer = getattr(config, "primary_viewer", None)
    else:
        viewer = settings.pop("viewer", getattr(config, "primary_viewer"))

    # Make sure viewer is visible
    if settings.show:
        viewer.show()

    # We need to pop clear/clear3d to prevent clearing again later
    if settings.clear:
        settings.clear = False  # clear only once
        viewer.clear()

    # Add object (the viewer currently takes care of producing the visuals)
    if neurons:
        # We need to pop viewer-specific settings to prevent errors in plotting functions
        neuron_settings = settings.to_dict()
        for key in settings._viewer_settings:
            neuron_settings.pop(key, None)
        viewer.add_neurons(neurons, center=settings.get("center", True), **neuron_settings)
    if volumes:
        for v in volumes:
            viewer.add_mesh(
                v,
                name=getattr(v, "name", None),
                color=getattr(v, "color", (0.95, 0.95, 0.95, 0.1)),
                alpha=getattr(v, "alpha", None),
                center=settings.center,
            )
    if points:
        for p in points:
            viewer.add_points(p, center=settings.center, **settings.scatter_kws)

    return viewer


def plot3d_plotly(x, **kwargs):
    """
    Plot3d() helper function to generate plotly 3D plots. This is just to
    improve readability and structure of the code.
    """
    # Lazy import because plotly is not a hard dependency
    try:
        import plotly.graph_objs as go
        from .plotly.graph_objs import (
            neuron2plotly,
            volume2plotly,
            scatter2plotly,
            layout2plotly,
        )
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "navis.plot3d() with the `plotly` backend requires the `plotly` library "
            "to be installed:\n  pip3 install plotly -U"
        )

    settings = PlotlySettings().update_settings(**kwargs)

    # Parse objects to plot
    (neurons, volumes, points, visual) = utils.parse_objects(x)

    neuron_cmap, volumes_cmap = prepare_colormap(
        settings.color,
        neurons=neurons,
        volumes=volumes,
        palette=settings.palette,
        color_by=None,
        alpha=settings.alpha,
        color_range=255,
    )

    data = []
    if neurons:
        data += neuron2plotly(neurons, neuron_cmap, settings)
    if volumes:
        data += volume2plotly(volumes, volumes_cmap, settings)
    if points:
        data += scatter2plotly(points, **settings.scatter_kws)

    layout = layout2plotly(**settings.to_dict())

    # If not provided generate a figure dictionary
    fig = settings.fig if settings.fig else go.Figure(layout=layout)
    if not isinstance(fig, (dict, go.Figure)):
        raise TypeError(
            "`fig` must be plotly.graph_objects.Figure or dict, got " f"{type(fig)}"
        )

    # Add data
    for trace in data:
        fig.add_trace(trace)

    if settings.inline and utils.is_jupyter():
        fig.show()
    else:
        logger.info("Use the `.show()` method to plot the figure.")
        return fig


def plot3d_k3d(x, **kwargs):
    """
    Plot3d() helper function to generate k3d 3D plots. This is just to
    improve readability and structure of the code.
    """
    # Lazy import because k3d is not (yet) a hard dependency
    try:
        import k3d
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "navis.plot3d() with `k3d` backend requires the k3d library "
            "to be installed:\n  pip3 install k3d -U"
        )

    from .k3d.k3d_objects import neuron2k3d, volume2k3d, scatter2k3d

    settings = K3dSettings().update_settings(**kwargs)

    # Parse objects to plot
    (neurons, volumes, points, visual) = utils.parse_objects(x)

    neuron_cmap, volumes_cmap = prepare_colormap(
        settings.color,
        neurons=neurons,
        volumes=volumes,
        palette=settings.palette,
        color_by=None,
        alpha=settings.alpha,
        color_range=255,
    )

    data = []
    if neurons:
        data += neuron2k3d(neurons, neuron_cmap, settings)
    if volumes:
        data += volume2k3d(volumes, volumes_cmap, settings)
    if points:
        data += scatter2k3d(points, **settings.scatter_kws)

    # If not provided generate a plot
    if not settings.plot:
        plot = k3d.plot(height=settings.height)
        plot.camera_rotate_speed = 5
        plot.camera_zoom_speed = 2
        plot.camera_pan_speed = 1
        plot.grid_visible = False

    # Add data
    for trace in data:
        plot += trace

    if settings.inline and utils.is_jupyter():
        plot.display()
    else:
        logger.info("Use the `.display()` method to show the plot.")
        return plot



import navis

import numpy as np


def test_parallel():
    # Load example neurons
    nl = navis.example_neurons(kind='skeleton')

    # Test decorator
    pr = navis.prune_by_strahler(nl, 1, parallel=True, inplace=False)

    assert isinstance(pr, navis.NeuronList)
    assert len(pr) == len(nl)
    assert pr[0].n_nodes < nl[0].n_nodes

    # Test apply
    pr = nl.apply(navis.prune_by_strahler, to_prune=1, inplace=False,
                  parallel=True)
    assert isinstance(pr, navis.NeuronList)
    assert len(pr) == len(nl)
    assert all(pr.n_nodes < nl.n_nodes)


def test_parallel_inplace():
    # Load example neurons
    nl = navis.example_neurons(kind='skeleton')

    # Test decorator with inplace=True
    pr = nl.copy()
    pr2 = navis.prune_by_strahler(pr, 1, parallel=True, inplace=True)
    assert len(pr) == len(pr2) == len(nl)
    assert pr[0].n_nodes == pr2[0].n_nodes

    # Test apply with inplace=True -> this should not work
    #pr = nl.apply(navis.prune_by_strahler, to_prune=1, inplace=True,
    #              parallel=True)
    #assert len(pr) == len(nl)
    #assert all(pr.n_nodes < nl.n_nodes)


def test_apply():
    # Load example neurons
    nl = navis.example_neurons(kind='skeleton')

    # Test apply
    ids = nl.apply(lambda x: x.id, parallel=False)
    assert isinstance(ids, list)
    assert len(ids) == len(nl)
    assert all(np.array(ids) == nl.id)

    # Test apply
    pr = nl.apply(navis.prune_by_strahler, to_prune=1, inplace=False,
                  parallel=False)
    assert isinstance(pr, navis.NeuronList)
    assert len(pr) == len(nl)
    assert all(pr.n_nodes < nl.n_nodes)

    # Test apply with inplace=True
    pr = nl.apply(navis.prune_by_strahler, to_prune=1, inplace=True,
                  parallel=False)
    assert isinstance(pr, navis.NeuronList)
    assert len(pr) == len(nl)
    assert all(pr.n_nodes == nl.n_nodes)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import datetime
import io
import os
import re
import requests
import tempfile
import tarfile

import multiprocessing as mp
import numpy as np
import pandas as pd

from abc import ABC
from functools import partial, wraps
from pathlib import Path
from typing import List, Union, Iterable, Dict, Optional, Any, IO
from typing_extensions import Literal
from zipfile import ZipFile, ZipInfo
from ftplib import FTP

from .. import config, utils, core

try:
    import zlib
    import zipfile

    compression = zipfile.ZIP_DEFLATED
except ModuleNotFoundError:
    compression = zipfile.ZIP_STORED

__all__ = ["BaseReader"]

# Set up logging
logger = config.get_logger(__name__)

DEFAULT_INCLUDE_SUBDIRS = False

# Regular expression to figure out if a string is a regex pattern
rgx = re.compile(r"[\\\.\?\[\]\+\^\$\*]")


def merge_dicts(*dicts: Optional[Dict], **kwargs) -> Dict:
    """Merge dicts and kwargs left to right.

    Ignores None arguments.
    """
    # handles nones
    out = dict()
    for d in dicts:
        if d:
            out.update(d)
    out.update(kwargs)
    return out


def handle_errors(func):
    """Decorator for read_buffer and read_dataframe methods to handle errors.

    Catches exceptions, logs/raises and potentially return `None`.

    Note: various other BaseReader methods have their own error handling.

    Parameters
    ----------
    func : callable
        Function to wrap.

    Returns
    -------
    callable
        Wrapped function.

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        attrs = kwargs.get("attrs", {})  # we rely on this being a keyword argument!
        try:
            return func(*args, **kwargs)
        except BaseException as e:
            # Check if we can provide any hint as to which file failed
            id = self.name_fallback
            for a in ("file", "origin", "name"):
                if a in attrs:
                    id = attrs[a]
                    break

            if self.errors == "raise":
                raise ReadError(
                    f"Error reading {id}. See above traceback for details."
                ) from e
            elif self.errors == "log":
                logger.exception(f"Failed to read {id}", exc_info=True)

            return None

    return wrapper


class Writer:
    """Writer class that takes care of things like filenames, archives, etc.

    Parameters
    ----------
    write_func :    callable
                    Writer function to write a single file to disk. Must
                    accept a `filepath` parameter.
    ext :           str, optional
                    File extension - e.g. '.swc'.

    """

    def __init__(self, write_func, ext):
        assert callable(write_func)
        if ext:
            assert isinstance(ext, str) and ext.startswith(".")
        self.write_func = write_func
        self.ext = ext

    def write_single(self, x, filepath, **kwargs):
        """Write single object to file."""
        # try to str.format any path-like
        try:
            as_str = os.fspath(filepath)
        except TypeError:
            raise ValueError(
                f'`filepath` must be str or pathlib.Path, got "{type(filepath)}"'
            )

        # Format filename (e.g. "{neuron.name}.swc")
        formatted_str = as_str.format(neuron=x)

        # If it was formatted, make sure it has correct extension
        if self.ext and formatted_str != as_str and not as_str.endswith(self.ext):
            raise ValueError(f"Formattable filepaths must end with '{self.ext}'")

        filepath = Path(formatted_str)

        # Expand user - otherwise .exists() might fail
        filepath = filepath.expanduser()

        # If not specified, generate filename
        if self.ext and not str(filepath).endswith(self.ext):
            filepath = filepath / f"{x.id}{self.ext}"

        # Make sure the parent directory exists
        if not filepath.parent.exists():
            raise ValueError(f"Parent folder {filepath.parent} must exist.")

        # Track the path we put this (and presumably all other files in)
        self.path = Path(filepath)
        while not self.path.is_dir():
            self.path = self.path.parent

        return self.write_func(x, filepath=filepath, **kwargs)

    def write_many(self, x, filepath, **kwargs):
        """Write multiple files to folder."""
        if not utils.is_iterable(filepath):
            # Assume this is a folder if it doesn't end with e.g. '.swc'
            is_filename = str(filepath).endswith(self.ext) if self.ext else False
            is_single = len(x) == 1
            is_formattable = "{" in str(filepath) and "}" in str(filepath)
            if not is_filename or is_single or is_formattable:
                filepath = [filepath] * len(x)
            else:
                raise ValueError(
                    "`filepath` must either be a folder, a "
                    "formattable filepath or a list of filepaths"
                    "when saving multiple neurons."
                )

        if len(filepath) != len(x):
            raise ValueError(
                f"Got {len(filepath)} file names for " f"{len(x)} neurons."
            )

        # At this point filepath is iterable
        filepath: Iterable[str]
        for n, f in config.tqdm(
            zip(x, filepath),
            disable=config.pbar_hide,
            leave=config.pbar_leave,
            total=len(x),
            desc="Writing",
        ):
            self.write_single(n, filepath=f, **kwargs)

    def write_zip(self, x, filepath, **kwargs):
        """Write files to zip."""
        filepath = Path(filepath).expanduser()
        # Parse pattern, if given
        pattern = "{neuron.id}" + (self.ext if self.ext else "")
        if "@" in str(filepath):
            pattern, filename = filepath.name.split("@")
            filepath = filepath.parent / filename

        # Make sure we have an iterable
        x = core.NeuronList(x)

        with ZipFile(filepath, mode="w") as zf:
            # Context-manager will remove temporary directory and its contents
            with tempfile.TemporaryDirectory() as tempdir:
                for n in config.tqdm(
                    x,
                    disable=config.pbar_hide,
                    leave=config.pbar_leave,
                    total=len(x),
                    desc="Writing",
                ):
                    # Save to temporary file
                    f = None
                    try:
                        # Generate temporary filename
                        f = os.path.join(tempdir, pattern.format(neuron=n))
                        # Write to temporary file
                        self.write_single(n, filepath=f, **kwargs)
                        # Add file to zip
                        zf.write(
                            f,
                            arcname=pattern.format(neuron=n),
                            compress_type=compression,
                        )
                    except BaseException:
                        raise
                    finally:
                        # Remove temporary file - we do this inside the loop
                        # to avoid unnecessarily occupying space as we write
                        if f:
                            os.remove(f)

        # Set filepath to zipfile -> this overwrite filepath set in write_single
        # (which would be the temporary file)
        self.path = Path(filepath)

    def write_any(self, x, filepath, **kwargs):
        """Write any to file. Default entry point."""
        # If target is a zipfile
        if isinstance(filepath, (str, Path)) and str(filepath).endswith(".zip"):
            return self.write_zip(x, filepath=filepath, **kwargs)
        elif isinstance(x, core.NeuronList):
            return self.write_many(x, filepath=filepath, **kwargs)
        else:
            return self.write_single(x, filepath=filepath, **kwargs)


class BaseReader(ABC):
    """Abstract reader to parse various inputs into neurons.

    Any subclass should implement at least one of `read_buffer` or
    `read_dataframe`. Entry methods such as `read_any` will pass
    and parse an input through to the appropriate method.

    Parameters
    ----------
    fmt :           str
                    A string describing how to parse filenames into neuron
                    properties. For example '{id}.swc'.
    file_ext :      str | tuple
                    The file extension(s) to look for when searching folders.
                    For example '.swc'. Alternatively, you can re-implement
                    the `is_valid_file` method for more complex filters. That
                    method needs to be able to deal with: Path objects, ZipInfo
                    objects and strings.
    name_fallback : str
                    Fallback for name when reading from e.g. string.
    attrs :         dict
                    Additional attributes to use when creating the neuron.
                    Will be overwritten by later additions (e.g. from `fmt`).
    ignore_hidden : bool
                    Whether to ignore files that start with "._".
    errors :        "raise" | "log" | "ignore"
                    What to do when an error is encountered:
                     - "raise" (default) will raise an error
                     - "log" will log a warning and return `None`
                     - "ignore" will return `None`
                    Applies only to errors in parsing file contents into neurons
                    not to errors in reading files, archives, URLs, etc.

    """

    def __init__(
        self,
        fmt: str,
        file_ext: str,
        name_fallback: str = "NA",
        read_binary: bool = False,
        attrs: Optional[Dict[str, Any]] = None,
        ignore_hidden=True,
        errors="raise",
    ):
        self.attrs = attrs
        self.fmt = fmt
        self.file_ext = file_ext
        self.name_fallback = name_fallback
        self.read_binary = read_binary
        self.ignore_hidden = ignore_hidden
        self.errors = errors

        assert errors in ("raise", "log", "ignore")

    @property
    def file_ext(self):
        return self._file_ext

    @file_ext.setter
    def file_ext(self, value):
        """Makes sure file_ext is always a tuple."""
        if isinstance(value, str):
            value = (value,)

        if any((ext.startswith("*") for ext in value)):
            raise ValueError('File extension must be ".ext", not "*.ext"')

        self._file_ext = value

    def format_output(self, x):
        """Format output into NeuronList.

        Replace this method if output is not (always) a NeuronList.
        See for example NrrdReader in nrrd_io.py.
        """
        if not x:
            return core.NeuronList([])
        else:
            return core.NeuronList([n for n in x if n])

    def files_in_dir(
        self, dpath: Path, include_subdirs: bool = DEFAULT_INCLUDE_SUBDIRS
    ) -> Iterable[Path]:
        """List files to read in directory."""
        if not isinstance(dpath, Path):
            dpath = Path(dpath)

        if "*" in str(dpath):
            pattern = str(dpath.name)
            dpath = dpath.parent
        else:
            pattern = "*"

        dpath = dpath.expanduser()
        if include_subdirs:
            pattern = os.path.join("**", pattern)

        yield from (f for f in dpath.glob(pattern) if self.is_valid_file(f))

    def is_valid_file(self, file):
        """Return true if file should be considered for reading."""
        if isinstance(file, ZipInfo):
            file = file.filename
        elif isinstance(file, tarfile.TarInfo):
            file = file.name
        elif isinstance(file, Path):
            file = file.name

        if self.ignore_hidden and str(file).startswith("._"):
            return False

        for ext in self.file_ext:
            if str(file).endswith(ext):
                return True
        return False

    def _make_attributes(
        self, *dicts: Optional[Dict[str, Any]], **kwargs
    ) -> Dict[str, Any]:
        """Combine attributes with a timestamp and those defined on the object.

        Later additions take precedence:

        - created_at (now)
        - object-defined attributes
        - additional dicts given as *args
        - additional attributes given as **kwargs

        Returns
        -------
        Dict[str, Any]
            Arbitrary string-keyed attributes.
        """
        return merge_dicts(
            dict(created_at=str(datetime.datetime.now())),
            self.attrs,
            *dicts,
            **kwargs,
        )

    def read_file_path(
        self, fpath: os.PathLike, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.BaseNeuron":
        """Read single file from path into a neuron.

        Parameters
        ----------
        fpath :     str | os.PathLike
                    Path to files.
        attrs :     dict or None
                    Arbitrary attributes to include in the neuron.

        Returns
        -------
        core.BaseNeuron
        """
        p = Path(fpath)
        with open(p, "rb" if self.read_binary else "r") as f:
            props = self.parse_filename(f.name)
            props["origin"] = str(p)
            return self.read_buffer(f, attrs=merge_dicts(props, attrs))

    def read_from_zip(
        self,
        files: Union[str, List[str]],
        zippath: os.PathLike,
        attrs: Optional[Dict[str, Any]] = None,
    ) -> "core.NeuronList":
        """Read given files from a zip into a NeuronList.

        Typically not used directly but via `read_zip()` dispatcher.

        Parameters
        ----------
        files :     zipfile.ZipInfo | list thereof
                    Files inside the ZIP file to read.
        zippath :   str | os.PathLike
                    Path to zip file.
        attrs :     dict or None
                    Arbitrary attributes to include in the TreeNeuron.

        Returns
        -------
        core.NeuronList

        """
        p = Path(zippath)
        files = utils.make_iterable(files)

        neurons = []
        with ZipFile(p, "r") as zip:
            for file in files:
                # Note the `file` is of type zipfile.ZipInfo here
                props = self.parse_filename(file.orig_filename)
                props["origin"] = str(p)
                try:
                    n = self.read_bytes(zip.read(file), attrs=merge_dicts(props, attrs))
                    neurons.append(n)
                except BaseException:
                    if self.errors == "ignore":
                        logger.warning(f'Failed to read "{file.filename}" from zip.')
                    else:
                        raise

        return self.format_output(neurons)

    def read_zip(
        self,
        fpath: os.PathLike,
        parallel="auto",
        limit: Optional[int] = None,
        attrs: Optional[Dict[str, Any]] = None,
    ) -> "core.NeuronList":
        """Read files from a zip into a NeuronList.

        This is a dispatcher for `.read_from_zip`.

        Parameters
        ----------
        fpath :     str | os.PathLike
                    Path to zip file.
        limit :     int, optional
                    Limit the number of files read from this directory.
        attrs :     dict or None
                    Arbitrary attributes to include in the TreeNeuron.

        Returns
        -------
        core.NeuronList

        """
        fpath = Path(fpath).expanduser()
        read_fn = partial(
            self.read_from_zip, zippath=fpath, attrs=attrs
        )
        neurons = parallel_read_archive(
            read_fn=read_fn,
            fpath=fpath,
            file_ext=self.is_valid_file,
            limit=limit,
            parallel=parallel,
        )
        return self.format_output(neurons)

    def read_tar(
        self,
        fpath: os.PathLike,
        limit: Optional[int] = None,
        attrs: Optional[Dict[str, Any]] = None,
        ignore_hidden: bool = True
    ) -> "core.NeuronList":
        """Read files from a tar archive into a NeuronList.

        Parameters
        ----------
        fpath :     str | os.PathLike
                    Path to tar file.
        limit :     int, optional
                    Limit the number of files read from this directory.
        attrs :     dict or None
                    Arbitrary attributes to include in the TreeNeuron.

        Returns
        -------
        core.NeuronList

        """
        p = Path(fpath).expanduser()
        file_ext = self.is_valid_file

        # Check the content of the tar file
        # N.B. the TarInfo objects are hashable but the hash changes
        # when the archive is re-opened. Therefore, we track the
        # filenames and not the TarInfo objects.
        to_read = []
        with tarfile.open(p, "r") as tf:
            for i, file in enumerate(tf):
                fpath = file.name  # full path inside the tar
                fname = fpath.split("/")[-1]  # just the filename
                if ignore_hidden and fname.startswith("._"):
                    continue
                if callable(file_ext):
                    if self.is_valid_file(file):
                        to_read.append(fpath)
                elif file_ext == "*":
                    to_read.append(fpath)
                elif file_ext and fname.endswith(file_ext):
                    to_read.append(fpath)
                elif "." not in file.filename:
                    to_read.append(fpath)

                if isinstance(limit, int) and i >= limit:
                    break

        if isinstance(limit, list):
            to_read = [f for f in to_read if f in limit]
        elif isinstance(limit, slice):
            to_read = to_read[limit]
        elif isinstance(limit, str):
            # Check if limit is a regex
            if rgx.search(limit):
                to_read = [f for f in to_read if re.search(limit, f.split("/")[-1])]
            else:
                to_read = [f for f in to_read if limit in f.split("/")[-1]]

        # Wrapper for progess bar
        prog = partial(
            config.tqdm,
            desc="Importing",
            total=len(to_read),
            disable=config.pbar_hide,
            leave=config.pbar_leave,
        )

        # N.B. tar.gz is a bunch of files concatenated and then compressed!
        # In consequence, random access is very slow because we may have to seek
        # through the whole archive to find the start of the requested file.
        # The workaround is to open the archive in streaming (e.g. "r|gz") mode,
        # iterate through the files in sequence and exract if the file is requested.
        # This is also why we are not using parallel processing here.
        # See also https://tinyurl.com/5n8wz54m (links to StackOverflow)
        neurons = []
        to_read = set(to_read)  # faster lookup
        with prog() as pbar:
            # Open the tar file in streaming mode with transparent compression
            with tarfile.open(p, "r|*") as tf:
                for t in tf:
                    # Skip files we don't want to read
                    if t.name not in to_read:
                        continue
                    try:
                        props = self.parse_filename(t.name.split("/")[-1])
                        props["origin"] = str(p)
                        n = self.read_bytes(
                            tf.extractfile(t).read(),
                            attrs=merge_dicts(props, attrs),
                        )
                        neurons.append(n)
                        to_read.remove(t.name)
                        pbar.update()
                    except BaseException as e:
                        if self.errors == "ignore":
                            logger.warning(f'Failed to read "{t.name}" from tar.')
                        else:
                            raise

                    # If we have read all (requested) files we can stop
                    if not len(to_read):
                        break

        return self.format_output(neurons)

    def read_ftp(
        self,
        url,
        parallel="auto",
        limit: Optional[int] = None,
        attrs: Optional[Dict[str, Any]] = None
    ) -> "core.NeuronList":
        """Read files from an FTP server.

        This is a dispatcher for `.read_from_ftp`.

        Parameters
        ----------
        url :       str
                    Can be the path to a single file or a directory.
        limit :     int, optional
                    Limit the number of files read from this directory.
        attrs :     dict or None
                    Arbitrary attributes to include in the TreeNeuron.

        Returns
        -------
        core.NeuronList

        """
        # Remove the ftp:// prefix
        url = url.replace("ftp://", "")

        # Split into server and path
        server, path = url.split("/", 1)

        # Check if server contains a port
        if ":" in server:
            server, port = server.split(":")
            port = int(port)
        else:
            port = 21  # default port

        read_fn = partial(self.read_from_ftp, attrs=attrs)
        neurons = parallel_read_ftp(
            read_fn=read_fn,
            server=server,
            port=port,
            path=path,
            file_ext=self.is_valid_file,
            limit=limit,
            parallel=parallel,
        )
        return self.format_output(neurons)

    def read_from_ftp(
        self,
        files: Union[str, List[str]],
        ftp: FTP,
        attrs: Optional[Dict[str, Any]] = None
    ) -> "core.NeuronList":
        """Read given files from an FTP server into a NeuronList.

        Typically not used directly but via `read_ftp()` dispatcher.

        Parameters
        ----------
        files :     tarfile.TarInfo | list thereof
                    Files inside the tar file to read.
        ftp :       ftplib.FTP | "GLOBAL"
                    The FTP client. This should already be connected, logged in
                    and in the correct directory. If "GLOBAL", we will look for a
                    `_FTP` global variable.
        attrs :     dict or None
                    Arbitrary attributes to include in the TreeNeuron.

        Returns
        -------
        core.NeuronList

        """
        # When reading in parallel, we expect there to be a global FTP connection
        # that was initialized once for each worker process.
        if ftp == "GLOBAL":
            if "_FTP" not in globals():
                raise ValueError("No global FTP connection found.")
            ftp = _FTP

        files = utils.make_iterable(files)

        neurons = []
        for file in files:
            # Read the file into a bytes
            with io.BytesIO() as f:
                ftp.retrbinary("RETR " + file, f.write)
                f.seek(0)
                props = self.parse_filename(file)
                props["origin"] = f"{ftp.host}:{ftp.port}{ftp.pwd()}/{file}"
                try:
                    n = self.read_buffer(f, attrs=merge_dicts(props, attrs))
                    neurons.append(n)
                except BaseException:
                    if self.errors == "ignore":
                        logger.warning(f'Failed to read "{file}" from FTP.')
                    else:
                        raise

        return self.format_output(neurons)

    def read_directory(
        self,
        path: os.PathLike,
        include_subdirs=DEFAULT_INCLUDE_SUBDIRS,
        parallel="auto",
        limit: Optional[int] = None,
        attrs: Optional[Dict[str, Any]] = None,
    ) -> "core.NeuronList":
        """Read directory of files into a NeuronList.

        Parameters
        ----------
        fpath :             str | os.PathLike
                            Path to directory containing files.
        include_subdirs :   bool, optional
                            Whether to descend into subdirectories, default False.
        parallel :          str | bool | "auto"
        limit :             int, optional
                            Limit the number of files read from this directory.
        attrs :             dict or None
                            Arbitrary attributes to include in the TreeNeurons
                            of the NeuronList

        Returns
        -------
        core.NeuronList

        """
        files = list(self.files_in_dir(Path(path), include_subdirs))

        if isinstance(limit, int):
            files = files[:limit]
        elif isinstance(limit, list):
            files = [f for f in files if f in limit]
        elif isinstance(limit, slice):
            files = files[limit]
        elif isinstance(limit, str):
            # Check if limit is a regex
            if rgx.search(limit):
                files = [f for f in files if re.search(limit, str(f.name))]
            else:
                files = [f for f in files if limit in str(f)]

        read_fn = partial(self.read_file_path, attrs=attrs)
        neurons = parallel_read(read_fn, files, parallel)
        return self.format_output(neurons)

    def read_url(
        self, url: str, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.BaseNeuron":
        """Read file from URL into a neuron.

        Parameters
        ----------
        url :       str
                    URL to file.
        attrs :     dict or None
                    Arbitrary attributes to include in the neuron.

        Returns
        -------
        core.BaseNeuron
        """
        # Note: originally, we used stream=True and passed `r.raw` to the
        # read_buffer function but that caused issue when there was more
        # than one chunk which would require us to concatenate the chunks
        # `via r.raw.iter_content()`.
        # Instead, we will simply read the whole content, wrap it in a BytesIO
        # and pass that to the read_buffer function. This is not ideal as it
        # will load the whole file into memory while the streaming solution
        # may have raised an exception earlier if the file was corrupted or
        # the wrong format.
        with requests.get(url, stream=False) as r:
            r.raise_for_status()
            props = self.parse_filename(url.split("/")[-1])
            props["origin"] = url
            return self.read_buffer(
                io.BytesIO(r.content), attrs=merge_dicts(props, attrs)
            )

    def read_string(
        self, s: str, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.BaseNeuron":
        """Read single string into a Neuron.

        Parameters
        ----------
        s :         str
                    String.
        attrs :     dict or None
                    Arbitrary attributes to include in the neuron.

        Returns
        -------
        core.BaseNeuron
        """
        sio = io.StringIO(s)
        return self.read_buffer(
            sio,
            attrs=merge_dicts({"name": self.name_fallback, "origin": "string"}, attrs),
        )

    def read_bytes(
        self, s: str, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.BaseNeuron":
        """Read bytes into a Neuron.

        Parameters
        ----------
        s :         bytes
                    Bytes.
        attrs :     dict or None
                    Arbitrary attributes to include in the neuron.

        Returns
        -------
        core.BaseNeuron
        """
        sio = io.BytesIO(s)
        return self.read_buffer(
            sio,
            attrs=merge_dicts({"name": self.name_fallback, "origin": "string"}, attrs),
        )

    @handle_errors
    def read_dataframe(
        self, nodes: pd.DataFrame, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.BaseNeuron":
        """Convert a DataFrame into a neuron.

        Parameters
        ----------
        nodes :     pandas.DataFrame
        attrs :     dict or None
                    Arbitrary attributes to include in the neuron.

        Returns
        -------
        core.BaseNeuron
        """
        raise NotImplementedError(
            "Reading DataFrames not implemented for " f"{type(self)}"
        )

    @handle_errors
    def read_buffer(
        self, f: IO, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.BaseNeuron":
        """Read buffer into a single neuron.



        Parameters
        ----------
        f :         IO
                    Readable buffer.
        attrs :     dict | None
                    Arbitrary attributes to include in the neuron.

        Returns
        -------
        core.NeuronObject
        """
        raise NotImplementedError(
            "Reading from buffer not implemented for " f"{type(self)}"
        )

    def read_any_single(
        self, obj, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.BaseNeuron":
        """Attempt to convert an arbitrary object into a neuron.

        Parameters
        ----------
        obj :       typing.IO | pandas.DataFrame | str | os.PathLike
                    Readable buffer, dataframe, path or URL to single file,
                    or parsable string.
        attrs :     dict or None
                    Arbitrary attributes to include in the neuron.

        Returns
        -------
        core.BaseNeuron
        """
        if hasattr(obj, "read"):
            return self.read_buffer(obj, attrs=attrs)
        if isinstance(obj, pd.DataFrame):
            return self.read_dataframe(obj, attrs=attrs)
        if isinstance(obj, os.PathLike):
            if str(obj).endswith(".zip"):
                return self.read_zip(obj, attrs=attrs)
            elif ".tar" in str(obj):
                return self.read_tar(obj, attrs=attrs)
            return self.read_file_path(obj, attrs=attrs)
        if isinstance(obj, str):
            # See if this might be a file (make sure to expand user)
            if os.path.isfile(os.path.expanduser(obj)):
                p = Path(obj).expanduser()
                if p.suffix == ".zip":
                    return self.read_zip(p, attrs=attrs)
                elif any(
                    str(p).endswith(f) for f in (".tar", "tar.gz", "tar.bz", "tar.bz2")
                ):
                    return self.read_tar(p, attrs=attrs)
                return self.read_file_path(p, attrs=attrs)
            if obj.startswith("http://") or obj.startswith("https://"):
                return self.read_url(obj, attrs=attrs)
            if obj.startswith("ftp://"):
                return self.read_ftp(obj, attrs=attrs)
            return self.read_string(obj, attrs=attrs)
        if isinstance(obj, bytes):
            return self.read_bytes(obj, attrs=attrs)
        raise ValueError(f"Could not read neuron from object of type '{type(obj)}'")

    def read_any_multi(
        self,
        objs,
        include_subdirs=DEFAULT_INCLUDE_SUBDIRS,
        parallel="auto",
        attrs: Optional[Dict[str, Any]] = None,
    ) -> "core.NeuronList":
        """Attempt to convert an arbitrary object into a NeuronList,
        potentially in parallel.

        Parameters
        ----------
        obj :               sequence
                            Sequence of anything readable by read_any_single or
                            directory path(s).
        include_subdirs :   bool
                            Whether to include subdirectories of a given
                            directory.
        parallel :          str | bool | int | None
                            "auto" or True for n_cores // 2, otherwise int for
                            number of jobs, or False for serial.
        attrs :             dict or None
                            Arbitrary attributes to include in each TreeNeuron
                            of the NeuronList.

        Returns
        -------
        core.NeuronList

        """
        if not utils.is_iterable(objs):
            objs = [objs]

        if not objs:
            logger.warning("No files found, returning empty NeuronList")
            return self.format_output(objs)

        new_objs = []
        for obj in objs:
            try:
                if is_dir(obj):
                    new_objs.extend(self.files_in_dir(obj, include_subdirs))
                    continue
            except TypeError:
                pass
            new_objs.append(obj)

        # `parallel` can be ("auto", threshold) in which case `threshold`
        # determines at what length we use parallel processing
        if isinstance(parallel, tuple):
            parallel, threshold = parallel
        else:
            threshold = 200

        if (
            isinstance(parallel, str)
            and parallel.lower() == "auto"
            and len(new_objs) < threshold
        ):
            parallel = False

        read_fn = partial(self.read_any_single, attrs=attrs)
        neurons = parallel_read(read_fn, new_objs, parallel)
        return self.format_output(neurons)

    def read_any(
        self,
        obj,
        include_subdirs=DEFAULT_INCLUDE_SUBDIRS,
        parallel="auto",
        limit=None,
        attrs: Optional[Dict[str, Any]] = None,
    ) -> "core.NeuronObject":
        """Attempt to read an arbitrary object into a neuron.

        Parameters
        ----------
        obj :       typing.IO | str | os.PathLike | pandas.DataFrame
                    Buffer, path to file or directory, URL, string, or
                    dataframe.

        Returns
        -------
        core.NeuronObject
        """
        if utils.is_iterable(obj) and not hasattr(obj, "read"):
            return self.read_any_multi(
                obj, parallel=parallel, include_subdirs=include_subdirs, attrs=attrs
            )
        else:
            try:
                if is_dir(obj):
                    return self.read_directory(
                        obj,
                        include_subdirs,
                        parallel=parallel,
                        limit=limit,
                        attrs=attrs,
                    )
            except TypeError:
                pass
            try:
                if os.path.isfile(os.path.expanduser(obj)) and str(obj).endswith(
                    ".zip"
                ):
                    return self.read_zip(
                        obj, parallel=parallel, limit=limit, attrs=attrs
                    )
                if os.path.isfile(os.path.expanduser(obj)) and ".tar" in str(obj):
                    return self.read_tar(obj, limit=limit, attrs=attrs)
                if isinstance(obj, str) and obj.startswith("ftp://"):
                    return self.read_ftp(
                        obj, parallel=parallel, limit=limit, attrs=attrs
                    )
            except TypeError:
                pass
            return self.read_any_single(obj, attrs=attrs)

    def parse_filename(self, filename: str) -> dict:
        """Extract properties from filename according to specified formatter.

        Parameters
        ----------
        filename : str

        Returns
        -------
        props :     dict
                    Properties extracted from filename.

        """
        # Make sure we are working with the filename not the whole path
        filename = Path(filename).name

        # Escape all special characters
        fmt = re.escape(self.fmt)

        # Unescape { and }
        fmt = fmt.replace("\\{", "{").replace("\\}", "}")

        # Replace all e.g. {name} with {.*}
        prop_names = []
        for prop in re.findall("{.*?}", fmt):
            prop_names.append(prop[1:-1].replace(" ", ""))
            fmt = fmt.replace(prop, "(.*)")

        # Match
        match = re.search(fmt, filename)

        if not match:
            raise ValueError(f'Unable to match "{self.fmt}" to filename "{filename}"')

        props = {"file": filename}
        for i, prop in enumerate(prop_names):
            for p in prop.split(","):
                # Ignore empty ("{}")
                if not p:
                    continue

                # If datatype was specified
                if ":" in p:
                    p, dt = p.split(":")

                    props[p] = match.group(i + 1)

                    if dt == "int":
                        props[p] = int(props[p])
                    elif dt == "float":
                        props[p] = float(props[p])
                    elif dt == "bool":
                        props[p] = bool(props[p])
                    elif dt == "str":
                        props[p] = str(props[p])
                    else:
                        raise ValueError(
                            f'Unable to interpret datatype "{dt}" ' f"for property {p}"
                        )
                else:
                    props[p] = match.group(i + 1)
        return props

    def _extract_connectors(self, nodes: pd.DataFrame) -> Optional[pd.DataFrame]:
        """Infer outgoing/incoming connectors from data.

        Parameters
        ----------
        nodes :     pd.DataFrame

        Returns
        -------
        Optional[pd.DataFrame]
                    With columns `["node_id", "x", "y", "z", "connector_id", "type"]`
        """
        return


class ImageReader(BaseReader):
    """Reader for image data."""

    def __init__(self, output, thin, threshold, dotprop_kwargs, **kwargs):
        super().__init__(**kwargs)
        self.output = output
        self.thin = thin
        self.threshold = threshold
        self.dotprop_kwargs = dotprop_kwargs

    def convert_image(self, data, attrs, header, voxdim, units, space_units):
        """Convert image data to desired output.

        Parameters
        ----------
        data :          np.ndarray
                        Image/Voxel data.
        attrs :         dict
                        Additional attributes to associate with the neuron.
        header :        dict
                        Header information.
        voxdim :        list of numbers
                        Voxel dimensions.
        units :         str | list, optional
                        Units for the neuron (e.g. "1um" or ["4um", "4um", "40um"]).
        space_units :   str, optional
                        Space units (e.g. "um").

        """
        if self.output == "dotprops":
            # If we're trying to get voxels from an image
            if data.ndim == 3:
                if self.threshold:
                    if self.threshold >= 1:
                        data = data >= self.threshold
                    elif self.threshold < 1 and self.threshold > 0:
                        data = data >= (self.threshold * data.max())
                    else:
                        raise ValueError(
                            "Threshold must be either >=1 or 0-1, got "
                            f"{self.threshold}"
                        )

                if self.thin:
                    from skimage.morphology import skeletonize

                    data = skeletonize(data)

                # Convert data to x/y/z coordinates
                # Note we need to multiply units before creating the Dotprops
                # - otherwise the KNN will be wrong
                x, y, z = np.where(data)
                points = np.vstack((x, y, z)).T
                points = points * voxdim

                if not len(points):
                    raise ValueError(
                        f"No points extracted from {self.name_fallback} file. Try lowering the threshold?"
                    )

                x = core.make_dotprops(points, **self.dotprop_kwargs)
            elif data.ndim == 2:
                if data.shape[1] == 3:
                    points, vect, alpha = data, None, None
                elif data.shape[1] == 6:
                    points, vect, alpha = data[:, :3], data[:, 3:6], None
                elif data.shape[1] == 7:
                    points, vect, alpha = data[:, :3], data[:, 3:6], data[:, 6]
                else:
                    raise ValueError(
                        "Expected data to be either (N, 3), (N, 6) "
                        f"or (N, 7) but {self.name_fallback} file contains {data.shape}"
                    )
                # Get `k` either from provided kwargs or the file's header
                k = self.dotprop_kwargs.pop("k", header.get("k", 20))

                x = core.Dotprops(
                    points, k=k, vect=vect, alpha=alpha, **self.dotprop_kwargs
                )
            else:
                raise ValueError(
                    "Data must be 2- or 3-dimensional to extract "
                    f"Dotprops, got {data.ndim}"
                )

            # Set units from space_units (points are already in physical space)
            if space_units:
                if isinstance(space_units, str):
                    x.units = f"1 {space_units}"
                elif len(space_units) == 3:
                    x.units = [f"1 {s}" for s in space_units]
        else:
            if data.ndim == 2:
                logger.warning(
                    f"Data in {self.name_fallback} file is of shape {data.shape} - "
                    "i.e. 2D. Could this be a point cloud or dotprops "
                    "instead of voxels?"
                )
            x = core.VoxelNeuron(data, units=units)

        # Header is special - we do not want to register it
        setattr(x, f"{self.name_fallback.lower()}_header", header)

        # Try adding properties one-by-one. If one fails, we'll keep track of it
        # in the `.meta` attribute
        meta = {}
        for k, v in attrs.items():
            try:
                x._register_attr(k, v)
            except (AttributeError, ValueError, TypeError):
                meta[k] = v

        if meta:
            x.meta = meta

        return x


def parallel_read(read_fn, objs, parallel="auto") -> List["core.NeuronList"]:
    """Read neurons from some objects with the given reader function,
    potentially in parallel.

    Reader function must be picklable.

    Parameters
    ----------
    read_fn :       Callable
    objs :          Iterable
    parallel :      str | bool | int
                    "auto" or True for `n_cores` // 2, otherwise int for number
                    of jobs, or false for serial.

    Returns
    -------
    core.NeuronList

    """
    try:
        length = len(objs)
    except TypeError:
        length = None

    prog = partial(
        config.tqdm,
        desc="Importing",
        total=length,
        disable=config.pbar_hide,
        leave=config.pbar_leave,
    )

    # `parallel` can be ("auto", threshold) in which case `threshold`
    # determines at what length we use parallel processing
    if isinstance(parallel, tuple):
        parallel, threshold = parallel
    else:
        threshold = 200

    if (
        isinstance(parallel, str)
        and parallel.lower() == "auto"
        and not isinstance(length, type(None))
        and length < threshold
    ):
        parallel = False

    if parallel:
        # Do not swap this as `isinstance(True, int)` returns `True`
        if isinstance(parallel, (bool, str)):
            n_cores = max(1, os.cpu_count() // 2)
        else:
            n_cores = int(parallel)

        with mp.Pool(processes=n_cores) as pool:
            results = pool.imap(read_fn, objs)
            neurons = list(prog(results))
    else:
        neurons = [read_fn(obj) for obj in prog(objs)]

    return neurons


def parallel_read_archive(
    read_fn,
    fpath,
    file_ext,
    reader=None,
    limit=None,
    parallel="auto",
    ignore_hidden=True,
) -> List["core.NeuronList"]:
    """Read neurons from a ZIP archive, potentially in parallel.

    Reader function must be picklable.

    Parameters
    ----------
    read_fn :       Callable
    fpath :         str | Path
    file_ext :      str | callable
                    File extension to search for - e.g. ".swc". `None` or `''`
                    are interpreted as looking for filenames without extension.
                    To include all files use `'*'`. Can also be callable that
                    accepts a filename and returns True or False depending on
                    if it should be included.
    limit :         int, optional
                    Limit the number of files read from this directory.
    parallel :      str | bool | int
                    "auto" or True for n_cores // 2, otherwise int for number of
                    jobs, or false for serial.
    ignore_hidden : bool
                    Archives zipped on OSX can end up containing a
                    `__MACOSX` folder with files that mirror the name of other
                    files. For example if there is a `123456.swc` in the archive
                    you might also find a `__MACOSX/._123456.swc`. Reading the
                    latter will result in an error. If ignore_hidden=True
                    we will simply ignore all file that starts with "._".

    Returns
    -------
    core.NeuronList

    """
    # Check zip content
    p = Path(fpath)
    to_read = []
    with ZipFile(p, "r") as zip:
        for i, file in enumerate(zip.filelist):
            fname = file.filename.split("/")[-1]
            if ignore_hidden and fname.startswith("._"):
                continue
            if callable(file_ext):
                if file_ext(file):
                    to_read.append(file)
            elif file_ext == "*":
                to_read.append(file)
            elif file_ext and fname.endswith(file_ext):
                to_read.append(file)
            elif "." not in file.filename:
                to_read.append(file)

            if isinstance(limit, int) and i >= limit:
                break

    if isinstance(limit, list):
        to_read = [f for f in to_read if f in limit]
    elif isinstance(limit, slice):
        to_read = to_read[limit]
    elif isinstance(limit, str):
        # Check if limit is a regex
        if rgx.search(limit):
            to_read = [f for f in to_read if re.search(limit, f.filename)]
        else:
            to_read = [f for f in to_read if limit in f.filename]

    prog = partial(
        config.tqdm,
        desc="Importing",
        total=len(to_read),
        disable=config.pbar_hide,
        leave=config.pbar_leave,
    )

    # `parallel` can be ("auto", threshold) in which case `threshold`
    # determines at what length we use parallel processing
    if isinstance(parallel, tuple):
        parallel, threshold = parallel
    else:
        threshold = 200

    if (
        isinstance(parallel, str)
        and parallel.lower() == "auto"
        and len(to_read) < threshold
    ):
        parallel = False

    if parallel:
        # Do not swap this as `isinstance(True, int)` returns `True`
        if isinstance(parallel, (bool, str)):
            n_cores = max(1, os.cpu_count() // 2)
        else:
            n_cores = int(parallel)

        with mp.Pool(processes=n_cores) as pool:
            results = pool.imap(read_fn, to_read)
            neurons = list(prog(results))
    else:
        neurons = [read_fn(obj) for obj in prog(to_read)]

    return neurons


def parallel_read_ftp(
    read_fn,
    server,
    port,
    path,
    file_ext,
    limit=None,
    parallel="auto",
) -> List["core.NeuronList"]:
    """Read neurons from an FTP server, potentially in parallel.

    Reader function must be picklable.

    Parameters
    ----------
    read_fn :       Callable
    server :        str
                    FTP server address.
    port :          int
                    FTP server port.
    path :          str
                    Path to directory containing files or single file.
    file_ext :      str | callable
                    File extension to search for - e.g. ".swc". `None` or `''`
                    are interpreted as looking for filenames without extension.
                    To include all files use `'*'`. Can also be callable that
                    accepts a filename and returns True or False depending on
                    if it should be included.
    limit :         int, optional
                    Limit the number of files read from this directory.
    parallel :      str | bool | int
                    "auto" or True for n_cores // 2, otherwise int for number of
                    jobs, or false for serial.

    Returns
    -------
    core.NeuronList

    """
    # Check if this is a single file
    is_single_file = False
    if "*" not in path:
        if isinstance(file_ext, str) and path.endswith(file_ext):
            is_single_file = True
        elif callable(file_ext) and file_ext(path.rsplit("/", 1)[1]):
            is_single_file = True

    if is_single_file:
        path, fname = path.rsplit("/", 1)
        to_read = [fname]
    else:
        pattern = ""
        # Check if path contains a "*." pattern - e.g. something like "*_raw.swc"
        if "*" in path:
            path, fname = path.rsplit("/", 1)
            pattern = fname

        # Remove leading /
        if path.startswith("/"):
            path = path[1:]

        # First check content
        with FTP() as ftp:
            ftp.connect(server, port)  # connect to server
            ftp.login()  # anonymous login
            ftp.cwd(path)  # change to path

            # Read content
            content = []
            ftp.retrlines(f"LIST {pattern}", content.append)

        # Parse content into filenames
        to_read = []
        for line in content:
            if not line:
                continue
            file = line.split()[-1].strip()

            if callable(file_ext):
                if file_ext(file):
                    to_read.append(file)
            elif file_ext == "*":
                to_read.append(file)
            elif file_ext and fname.endswith(file_ext):
                to_read.append(file)

    if isinstance(limit, int):
        to_read = to_read[:limit]
    elif isinstance(limit, list):
        to_read = [f for f in to_read if f in limit]
    elif isinstance(limit, slice):
        to_read = to_read[limit]
    elif isinstance(limit, str):
        # Check if limit is a regex
        if rgx.search(limit):
            to_read = [f for f in to_read if re.search(limit, f)]
        else:
            to_read = [f for f in to_read if limit in f]

    if not to_read:
        return []

    prog = partial(
        config.tqdm,
        desc="Loading",
        total=len(to_read),
        disable=config.pbar_hide,
        leave=config.pbar_leave,
    )

    # `parallel` can be ("auto", threshold) in which case `threshold`
    # determines at what length we use parallel processing
    if isinstance(parallel, tuple):
        parallel, threshold = parallel
    else:
        threshold = 200

    if (
        isinstance(parallel, str)
        and parallel.lower() == "auto"
        and len(to_read) < threshold
    ):
        parallel = False

    if parallel:
        # Do not swap this as `isinstance(True, int)` returns `True`
        if isinstance(parallel, (bool, str)):
            n_cores = max(1, os.cpu_count() // 2)
        else:
            n_cores = int(parallel)

        # We can't send the FTP object to the process (because its socket is not pickleable)
        # Instead, we need to initialize a new FTP connection in each process via a global variable
        with mp.Pool(
            processes=n_cores, initializer=_ftp_pool_init, initargs=(server, port, path)
        ) as pool:
            results = pool.imap(partial(read_fn, ftp="GLOBAL"), to_read)
            neurons = list(prog(results))
    else:
        with FTP() as ftp:
            ftp.connect(server, port)
            ftp.login()
            ftp.cwd(path)

            neurons = [read_fn(file, ftp=ftp) for file in prog(to_read)]

    return neurons


def _ftp_pool_init(server, port, path):
    global _FTP
    _FTP = FTP()
    _FTP.connect(server, port)
    _FTP.login()
    _FTP.cwd(path)


def parse_precision(precision: Optional[int]):
    """Convert bit width into int and float dtypes.

    Parameters
    ----------
    precision : int
        16, 32, 64, or None

    Returns
    -------
    tuple
        Integer numpy dtype, float numpy dtype

    Raises
    ------
    ValueError
        Unknown precision.
    """
    INT_DTYPES = {16: np.int16, 32: np.int32, 64: np.int64, None: None}
    FLOAT_DTYPES = {16: np.float16, 32: np.float32, 64: np.float64, None: None}

    try:
        return (INT_DTYPES[precision], FLOAT_DTYPES[precision])
    except KeyError:
        raise ValueError(
            f"Unknown precision {precision}. Expected on of the following: 16, 32 (default), 64 or None"
        )


class ReadError(Exception):
    """Error raised when reading a file fails."""


def is_dir(path: os.PathLike) -> bool:
    """Check if path is a directory.

    The main purpose of this function is to catch
    *.file_ext at the end of the path.

    Parameters
    ----------
    path : os.PathLike
        Path to check.

    Returns
    -------
    bool
        True if path is a directory.

    """
    # Expand user
    path = os.path.expanduser(path)

    # Remove any trailing *.file_ext
    path = path.split("*")[0]

    return os.path.isdir(path)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import numpy as np
import trimesh as tm

from .. import core, utils


def simplify_mesh_fqmr(x, F, inplace=False, **kwargs):
    """Simplify mesh using pyfqmr.

    Parameters
    ----------
    x :         MeshNeuron | Volume | Trimesh
                Mesh object to simplify.
    F :         int
                Target face count (integer).
    inplace :   bool
                If True, will perform simplication on `x`. If False, will
                simplify and return a copy.
    **kwargs
                Keyword arguments are passed through to pyfqmr's
                `pyfqmr.Simplify.simplify_mesh`.

    Returns
    -------
    simp
                Simplified mesh object.

    """
    if not utils.is_mesh(x):
        raise TypeError(f'Expected mesh-like, got "{type(x)}"')

    try:
        import pyfqmr
    except ModuleNotFoundError:
        raise ModuleNotFoundError('Please install pyfqmr: pip3 install pyfqmr')

    defaults = dict(aggressiveness=7, preserve_border=True, verbose=False)
    defaults.update(kwargs)

    mesh_simplifier = pyfqmr.Simplify()
    mesh_simplifier.setMesh(x.vertices, x.faces)
    mesh_simplifier.simplify_mesh(target_count=F, **defaults)
    vertices, faces, normals = mesh_simplifier.getMesh()

    if not inplace:
        x = x.copy()

    x.vertices = vertices
    x.faces = faces

    if hasattr(x, 'face_normals'):
        x.face_normals = normals

    return x



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.


""" This module contains functions to analyse neuron's form factors."""

import os
import scipy

import pandas as pd
import multiprocessing as mp
import numpy as np

from functools import partial
from typing import Union, Optional, Sequence, List, Dict, overload
from typing_extensions import Literal

from .. import config, core, utils

# Set up logging
logger = config.get_logger(__name__)

__all__ = sorted(['form_factor'])


def form_factor(x: Union['core.TreeNeuron', 'core.MeshNeuron'],
                start: int = -3,
                stop: int = 3,
                num: int = 601,
                parallel: bool = False,
                n_cores: int = os.cpu_count() // 2,
                progress=True):
    """Calculate form factor for given neuron.

    The form factor F(q) is a Fourier transform of density-density correlation
    of particles used to classify objects in polymer physics. Based on Choi et
    al., 2022 (bioRxiv). Code adapted from github.com/kirichoi/FqClustering.

    Parameters
    ----------
    x :         TreeNeuron | Meshneuron | Dotprops | NeuronList
                Neurons to calculate form factor for. A few notes:
                  - data should be in micron - if not, you might want to adjust
                    start/stop/min!
                  - since this is all about density, it may make sense to
                    resample neurons
    start/stop/num : int
                Start/stop/num describe the (log) space over which to calculate
                the form factor. Effectively determining the resolution.
                Assuming `x` is in microns the defaults mean we pay attention
                to densities between 1 nm (1e-3 microns) and 1 mm (1e+3 microns).
                The x-value corresponding to the form factor(s) in `Fq` will
                be `np.logspace(start, stop, num)`.
    parallel :  bool
                Whether to use multiple cores when `x` is a NeuronList.
    n_cores :   bool
                Number of cores to use when `x` is a NeuronList and
                `parallel=True`. Even on a single core this function makes
                heavy use of numpy which itself uses multiple threads - it is
                therefore not advisable to use all your cores as this would
                create a bottleneck.
    progress :  bool
                Whether to show a progress bar.

    Returns
    -------
    Fq :        np.ndarray
                For single neurons: `(num,)` array
                For Neuronlists: `(len(x), num)` array

    References
    ----------
    Polymer physics-based classification of neurons
    Kiri Choi, Won Kyu Kim, Changbong Hyeon
    bioRxiv 2022.04.07.487455; doi: https://doi.org/10.1101/2022.04.07.487455

    Examples
    --------
    >>> import navis
    >>> nl = navis.example_neurons(3)
    >>> nl = nl.convert_units('microns')
    >>> # Resample to 1 node / micron
    >>> rs = navis.resample_skeleton(nl, '1 micron')
    >>> # Calculate form factor
    >>> Fq = navis.form_factor(rs, start=-3, stop=3, num=301,
    ...                        parallel=True, n_cores=3)
    >>> # Plot
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> x = np.logspace(-3, 3,  301)
    >>> fig, ax = plt.subplots()
    >>> for i in range(len(Fq)):
    ...     _ = ax.plot(x, Fq[i])
    >>> # Make log-log
    >>> ax.set_xscale('log')
    >>> ax.set_yscale('log')
    >>> plt.show()                                              # doctest: +SKIP
    >>> # Cluster
    >>> from scipy.spatial.distance import pdist
    >>> from scipy.cluster.hierarchy import dendrogram, linkage
    >>> dists = pdist(Fq)
    >>> Z = linkage(dists, method='ward')
    >>> dn = dendrogram(Z)                                      # doctest: +SKIP

    """
    if isinstance(x, core.NeuronList):
        pbar = partial(
            config.tqdm,
            desc='Calc. form factor',
            total=len(x),
            disable=config.pbar_hide or not progress,
            leave=config.pbar_leave
        )
        _calc_form_factor = partial(form_factor, progress=False,
                                    start=start, stop=stop, num=num)

        if parallel:
            with mp.Pool(processes=n_cores) as pool:
                results = pool.imap(_calc_form_factor, x)
                Fq = list(pbar(results))
        else:
            Fq = [_calc_form_factor(n) for n in pbar(x)]

        return np.vstack(Fq)

    utils.eval_param(x, name='x', allowed_types=(core.TreeNeuron,
                                                 core.Dotprops,
                                                 core.MeshNeuron))

    if isinstance(x, core.TreeNeuron):
        coor = x.nodes[['x', 'y', 'z']].values
    elif isinstance(x, core.MeshNeuron):
        coor = x.vertices
    elif isinstance(x, core.Dotprops):
        coor = x.points

    ucoor = np.unique(coor, axis=0)
    lenucoor = len(ucoor)

    q_range = np.logspace(start, stop, num)
    Fq = np.empty(len(q_range))
    ccdisttri = scipy.spatial.distance.pdist(ucoor)

    for q in config.trange(len(q_range),
                           desc='Calc. form factor',
                           disable=config.pbar_hide or not progress,
                           leave=config.pbar_leave):
        qrvec = q_range[q] * ccdisttri
        Fq[q] = np.divide(np.divide(2 * np.sum(np.sin(qrvec) / qrvec), lenucoor), lenucoor) + 1 / lenucoor

    return Fq



from __future__ import annotations
from abc import ABC, abstractmethod
from itertools import permutations
import sys
import os
from collections import Counter
from concurrent.futures import ProcessPoolExecutor
from typing import (
    Generic,
    Hashable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Callable,
    List,
    Iterable,
    Any,
    Tuple,
    TypeVar,
    Union,
)
import logging
from pathlib import Path
from functools import lru_cache
from copy import deepcopy
import operator
import math
from collections import defaultdict

import numpy as np
import pandas as pd

from .. import config

# from ..core import Dotprops
from .. import core

logger = logging.getLogger(__name__)

DEFAULT_SEED = 1991

epsilon = sys.float_info.epsilon
cpu_count = max(1, os.cpu_count() - 1)

fp = Path(__file__).resolve().parent
smat_path = fp / "score_mats"


def chunksize(it_len, cpu_count, min_chunk=50):
    return max(min_chunk, int(it_len / (cpu_count * 4)))


def yield_not_same(pairs: Iterable[Tuple[Any, Any]]) -> Iterator[Tuple[Any, Any]]:
    for a, b in pairs:
        if a != b:
            yield a, b


def concat_results(
    results: Iterable[List[np.ndarray]],
    total: Optional[int] = None,
    desc: str = "Querying",
    progress: bool = True,
) -> List[np.ndarray]:
    """Helper function to concatenate batches of e.g. [(dist, dots), (dist, dots)]
    into single (dist, dot) arrays.
    """
    intermediate = defaultdict(list)
    with config.tqdm(desc=desc, total=total, leave=False, disable=not progress) as pbar:
        for result_lst in results:
            for idx, array in enumerate(result_lst):
                intermediate[idx].append(array)
            pbar.update(1)

    return [np.concatenate(arrs) for arrs in intermediate.values()]


def _nblast_v1_scoring(dist: float, dp: float, sigma_scoring: int = 10):
    """NBLAST analytical scoring function following Kohl et al. (2013).

    Parameters
    ----------
    dist :          float | array thereof
                    Distance between two points.
    dp :            float | array thereof
                    Absolute dot product between points.
    sigma_scoring : int
                    Sigma of the exponential decrease.
                    It determines how close in space points must be to be
                    considered similar. Defaults to 10.
    Returns
    -------
    scores :        float or array thereof
                    Score value(s).

    References
    ----------
    Kohl J, Ostrovsky AD, Frechter S, Jefferis GS. A bidirectional circuit switch
    reroutes pheromone signals in male and female brains.
    Cell. 2013 Dec;155(7) 1610-1623. doi: 10.1016/j.cell.2013.11.025.
    """
    return np.sqrt(np.abs(dp) * np.exp(-(dist**2) / (2 * sigma_scoring**2)))


NeuronKey = Hashable
T = TypeVar("T")


class LookupNdBuilder:
    """Class for building an N-dimensional score lookup (for e.g. NBLAST).

    Once instantiated, the axes of the lookup table must be defined.
    Call `.with_digitizers()` to manually define them, or
    `.with_bin_counts()` to learn them from the matched-pair data.

    Then call `.build()` to build the lookup table.

    Parameters
    ----------
    neurons :       dict or list of objects (e.g. Dotprops)
                    An indexable, consistently-ordered sequence of all
                    objects (typically neurons) which will be used as the
                    training set. Importantly: each object must have a
                    `len()`!
    matching_sets : list of lists of index into `neurons`
                    Lists of neurons, as indices into `neurons`, which
                    should be considered matches:

                        [[0, 1, 2, 4], [5, 6], [9, 10, 11]]

    match_fn :      Callable[[object, object], List[np.ndarray[float]]]
                    Function taking 2 arguments, both instances of type
                    `neurons`, and returning a list of 1D
                    `numpy.ndarray`s of floats. The length of the list
                    must be the same as the length of `boundaries`.
                    The length of the `array`s must be the same as the
                    number of points in the first argument. This function
                    returns values describing the quality of point matches
                    from a query to a target neuron.
    nonmatching :   list of index into `neurons`, optional
                    List of objects, as indices into `neurons`, which
                    should be be considered NON-matches. If not given,
                    all `neurons` will be used (on the assumption that
                    matches are a small subset of possible pairs).
    draw_strat :    "batched" | "greedy"
                    Strategy for randomly drawing non-matching pairs. Only
                    relevant if `nonmatching` is not provided.
                    "batched" should be the right choice in most scenarios.
                    "greedy" can be better if your pool of neurons is very
                    small.
    seed :          int, optional
                    Non-matching pairs are drawn at random using this seed,
                    by default {DEFAULT_SEED}.

    """

    def __init__(
        self,
        neurons: Union[List[T], Mapping[NeuronKey, T]],
        matching_lists: List[List[NeuronKey]],
        match_fn: Callable[[T, T], List[np.ndarray]],
        nonmatching_list: Optional[List[NeuronKey]] = None,
        draw_strat: str = "batched",
        seed: int = DEFAULT_SEED,
    ) -> None:
        self.objects = neurons
        self.matching_lists = matching_lists
        self._nonmatching_list = nonmatching_list
        self.match_fn = match_fn
        self.nonmatching_draw = draw_strat

        self.digitizers: Optional[List[Digitizer]] = None
        self.bin_counts: Optional[List[int]] = None

        self.seed = seed
        self._ndim: Optional[int] = None

    @property
    def ndim(self) -> int:
        if self._ndim is None:
            idx1, idx2 = self._object_keys()[:2]
            self._ndim = len(self._query(idx1, idx2))
        return self._ndim

    def with_digitizers(self, digitizers: List[Digitizer]):
        """Specify the axes of the output lookup table directly.

        Parameters
        ----------
        digitizers : List[Digitizer]

        Returns
        -------
        self
            For chaining convenience.
        """
        if len(digitizers) != self.ndim:
            raise ValueError(
                f"Match function returns {self.ndim} values "
                f"but provided {len(digitizers)} digitizers"
            )

        self.digitizers = digitizers
        self.bin_counts = None
        return self

    def with_bin_counts(self, bin_counts: List[int], method="quantile"):
        """Specify the number of bins on each axis of the output lookup table.

        The bin boundaries will be determined by evenly partitioning the data
        from the matched pairs into quantiles, in each dimension.

        Parameters
        ----------
        bin_counts : List[int]
        method :     'quantile' | 'geometric' | 'linear'
                     Method used to tile the data space.

        Returns
        -------
        self
            For chaining convenience.
        """
        if len(bin_counts) != self.ndim:
            raise ValueError(
                f"Match function returns {self.ndim} values "
                f"but provided {len(bin_counts)} bin counts"
            )

        self.bin_counts = bin_counts
        self.digitizers = None
        self.bin_method = method
        return self

    def _object_keys(self) -> Sequence[NeuronKey]:
        """Get all indices into objects instance member."""
        try:
            return self.objects.keys()
        except AttributeError:
            return range(len(self.objects))

    @property
    def nonmatching(self) -> List[NeuronKey]:
        """Indices of nonmatching set of neurons."""
        if self._nonmatching_list is None:
            return list(self._object_keys())
        return self._nonmatching_list

    def _yield_matching_pairs(self) -> Iterator[Tuple[NeuronKey, NeuronKey]]:
        """Yield all index pairs within all matching pairs."""
        for ms in self.matching_lists:
            yield from yield_not_same(permutations(ms, 2))

    def _yield_nonmatching_pairs(self) -> Iterator[Tuple[NeuronKey, NeuronKey]]:
        """Yield all index pairs within all non-matching pairs."""
        if self._nonmatching_list is None:
            raise ValueError("Must provide non-matching pairs explicitly.")
        yield from yield_not_same(permutations(self._nonmatching_list, 2))

    def _yield_nonmatching_pairs_greedy(
        self, rng=None
    ) -> Iterator[Tuple[NeuronKey, NeuronKey]]:
        """Yield all index pairs within nonmatching list."""
        return yield_not_same(permutations(self.nonmatching, 2))

    def _yield_nonmatching_pairs_batched(self) -> Iterator[Tuple[NeuronKey, NeuronKey]]:
        """Yield all index pairs within nonmatching list.

        This function tries to generate truely random draws of all possible
        non-matching pairs without actually having to generate all pairs.
        Instead, we generate new randomly permutated pairs in batches from which
        we then remove previously seen pairs.

        This works reasonable well as long as we only need a small subset
        of all possible non-matches. Otherwise this becomes inefficient.

        """
        nonmatching = np.array(self.nonmatching)

        seen = []
        rng = np.random.default_rng(self.seed)

        # Generate random pairs
        pairs = np.vstack(
            (rng.permutation(nonmatching), rng.permutation(nonmatching))
        ).T
        pairs = pairs[pairs[:, 0] != pairs[:, 1]]  # drop self hits
        seen = set([tuple(p) for p in pairs])  # track already seen pairs
        i = 0
        while True:
            # If exhausted, generate a new batch of random permutation
            if i >= len(pairs):
                pairs = np.vstack(
                    (rng.permutation(nonmatching), rng.permutation(nonmatching))
                ).T
                pairs = pairs[pairs[:, 0] != pairs[:, 1]]  # drop self hits
                pairs = set([tuple(p) for p in pairs])
                pairs = pairs - seen
                seen = seen | pairs
                pairs = list(pairs)
                i = 0

            # Pick a pair
            ix1, ix2 = pairs[i]
            i += 1

            yield (ix1, ix2)

    def _empty_counts(self) -> np.ndarray:
        """Create an empty array in which to store counts; shape determined by digitizer sizes."""
        shape = [len(b) for b in self.digitizers]
        return np.zeros(shape, int)

    def _query(self, q_idx, t_idx) -> List[np.ndarray]:
        """Get the results of applying the match function to objects specified by indices."""
        return self.match_fn(self.objects[q_idx], self.objects[t_idx])

    def _query_many(self, idx_pairs, threads=None) -> Iterator[List[np.ndarray]]:
        """Yield results from querying many pairs of neuron indices."""
        if threads is None or (threads == 0 and cpu_count == 1):
            for q_idx, t_idx in idx_pairs:
                yield self._query(q_idx, t_idx)
            return

        threads = threads or cpu_count
        idx_pairs = np.asarray(idx_pairs)
        chunks = chunksize(len(idx_pairs), threads)

        with ProcessPoolExecutor(threads) as exe:
            yield from exe.map(
                self.match_fn,
                [self.objects[ix] for ix in idx_pairs[:, 0]],
                [self.objects[ix] for ix in idx_pairs[:, 1]],
                chunksize=chunks,
            )

    def _query_to_idxs(self, q_idx, t_idx, counts=None):
        """Produce a digitized counts array from a given query-target pair."""
        return self._count_results(self._query(q_idx, t_idx), counts)

    def _count_results(self, results: List[np.ndarray], counts=None):
        """Convert raw match function ouput into a digitized counts array.

        Requires digitizers.
        """
        # Digitize
        idxs = [dig(r) for dig, r in zip(self.digitizers, results)]

        # Make a stack
        stack = np.vstack(idxs).T

        # Create empty matrix if necessary
        if counts is None:
            counts = self._empty_counts()

        # Get counts per cell -> this is the actual bottleneck of this function
        cells, cnt = np.unique(stack, axis=0, return_counts=True)

        # Fill matrix
        counts[tuple(cells[:, i] for i in range(cells.shape[1]))] += cnt

        return counts

    def _counts_array(
        self,
        idx_pairs,
        threads=None,
        progress=True,
        desc=None,
    ):
        """Convert index pairs into a digitized counts array.

        Requires digitizers.
        """
        counts = self._empty_counts()
        if threads is None or (threads == 0 and cpu_count == 1):
            for q_idx, t_idx in config.tqdm(
                idx_pairs, leave=False, desc=desc, disable=not progress
            ):
                counts = self._query_to_idxs(q_idx, t_idx, counts)
            return counts

        threads = threads or cpu_count
        idx_pairs = np.asarray(idx_pairs, dtype=int)
        chunks = chunksize(len(idx_pairs), threads)

        # because digitizing is not necessarily free,
        # keep this parallelisation separate to that in _query_many
        with ProcessPoolExecutor(threads) as exe:
            # This is the progress bar
            with config.tqdm(
                desc=desc, total=len(idx_pairs), leave=False, disable=not progress
            ) as pbar:
                for distdots in exe.map(
                    self.match_fn,
                    [self.objects[ix] for ix in idx_pairs[:, 0]],
                    [self.objects[ix] for ix in idx_pairs[:, 1]],
                    chunksize=chunks,
                ):
                    counts = self._count_results(distdots, counts)
                    pbar.update(1)

        return counts

    def _pick_nonmatching_pairs(self, n_matching_qual_vals, progress=True):
        """Using the seeded RNG, pick which non-matching pairs to use."""
        # pre-calculating which pairs we're going to use,
        # rather than drawing them as we need them,
        # means that we can parallelise the later step more effectively.
        # Slowdowns here are practically meaningless
        # because of how long distdot calculation will take
        nonmatching_pairs = []
        n_nonmatching_qual_vals = 0
        if self.nonmatching_draw == "batched":
            # This is a generator that tries to generate random pairs in
            # batches to avoid having to calculate all possible pairs
            gen = self._yield_nonmatching_pairs_batched()
            with config.tqdm(
                desc="Drawing non-matching pairs",
                total=n_matching_qual_vals,
                leave=False,
                disable=not progress,
            ) as pbar:
                # Draw non-matching pairs until we have enough data
                for nonmatching_pair in gen:
                    nonmatching_pairs.append(nonmatching_pair)
                    new_vals = len(self.objects[nonmatching_pair[0]])
                    n_nonmatching_qual_vals += new_vals

                    pbar.update(new_vals)

                    if n_nonmatching_qual_vals >= n_matching_qual_vals:
                        break
        elif self.nonmatching_draw == "greedy":
            # Generate all possible non-matching pairs
            possible_pairs = len(self.nonmatching) ** 2 - len(self.nonmatching)
            all_nonmatching_pairs = [
                p
                for p in config.tqdm(
                    self._yield_nonmatching_pairs_greedy(),
                    total=possible_pairs,
                    desc="Generating non-matching pairs",
                )
            ]
            # Randomly pick non-matching pairs until we have enough data
            rng = np.random.default_rng(self.seed)
            with config.tqdm(
                desc="Drawing non-matching pairs",
                total=n_matching_qual_vals,
                leave=False,
                disable=not progress,
            ) as pbar:
                while n_nonmatching_qual_vals < n_matching_qual_vals:
                    idx = rng.integers(0, len(all_nonmatching_pairs))
                    nonmatching_pair = all_nonmatching_pairs.pop(idx)
                    nonmatching_pairs.append(nonmatching_pair)

                    new_vals = len(self.objects[nonmatching_pair[0]])
                    n_nonmatching_qual_vals += new_vals

                    pbar.update(new_vals)
        else:
            raise ValueError(
                "Unknown strategy for non-matching pair draw:"
                f"{self.nonmatching_draw}"
            )

        return nonmatching_pairs

    def _get_pairs(self):
        matching_pairs = list(set(self._yield_matching_pairs()))

        # If no explicit non-matches provided, pick them from the entire pool
        if self._nonmatching_list is None:
            # need to know the eventual distdot count
            # so we know how many non-matching pairs to draw
            q_idx_count = Counter(p[0] for p in matching_pairs)
            n_matching_qual_vals = sum(
                len(self.objects[q_idx]) * n_reps
                for q_idx, n_reps in q_idx_count.items()
            )

            nonmatching_pairs = self._pick_nonmatching_pairs(n_matching_qual_vals)
        else:
            nonmatching_pairs = list(set(self._yield_nonmatching_pairs()))

        return matching_pairs, nonmatching_pairs

    def _build(self, threads, progress=True) -> Tuple[List[Digitizer], np.ndarray]:
        # Asking for more threads than available CPUs seems to crash on Github
        # actions
        if threads and threads >= cpu_count:
            threads = cpu_count

        if self.digitizers is None and self.bin_counts is None:
            raise ValueError(
                "Builder needs either digitizers or bin_counts - " "see with_* methods."
            )

        self.matching_pairs, self.nonmatching_pairs = self._get_pairs()

        logger.info("Comparing matching pairs")
        if self.digitizers:
            self.match_counts_ = self._counts_array(
                self.matching_pairs,
                threads=threads,
                progress=progress,
                desc="Comparing matching pairs",
            )
        else:
            match_results = concat_results(
                self._query_many(self.matching_pairs, threads),
                progress=progress,
                desc="Comparing matching pairs",
                total=len(self.matching_pairs),
            )
            self.match_results_ = match_results
            self.digitizers = []
            for i, (data, nbins) in enumerate(zip(match_results, self.bin_counts)):
                if not isinstance(nbins, Digitizer):
                    try:
                        self.digitizers.append(
                            Digitizer.from_data(data, nbins, method=self.bin_method)
                        )
                    except BaseException as e:
                        logger.error(f"Error creating digitizers for axes {i + 1}")
                        raise e
                else:
                    self.digitizers.append(nbins)

            logger.info("Counting results (this may take a while)")
            self.match_counts_ = self._count_results(match_results)

        logger.info("Comparing non-matching pairs")
        self.nonmatch_counts_ = self._counts_array(
            self.nonmatching_pairs,
            threads=threads,
            progress=progress,
            desc="Comparing non-matching pairs",
        )

        # Account for there being different total numbers of datapoints for
        # matches and nonmatches
        self.matching_factor_ = self.nonmatch_counts_.sum() / self.match_counts_.sum()
        if np.any(self.match_counts_ + self.nonmatch_counts_ == 0):
            logger.warning("Some lookup cells have no data in them")

        self.cells_ = np.log2(
            (self.match_counts_ * self.matching_factor_ + epsilon)
            / (self.nonmatch_counts_ + epsilon)
        )

        return self.digitizers, self.cells_

    def build(self, threads=None) -> LookupNd:
        """Build the score matrix.

        All non-identical neuron pairs within all matching sets are selected,
        and the scoring function is evaluated for those pairs.
        Then, the minimum number of non-matching pairs are randomly drawn
        so that at least as many data points can be calculated for non-matching
        pairs.

        In each bin of the score matrix, the log2 odds ratio of a score
        in that bin belonging to a match vs. non-match is calculated.

        Parameters
        ----------
        threads :   int, optional
                    If None, act in serial.
                    If 0, use cpu_count - 1.
                    Otherwise, use the given value.
                    Will be clipped at number of available cores - 1.
                    Note that with the currently implementation a large number
                    of threads might (and somewhat counterintuitively) actually
                    be slower than running building the scoring function in serial.

        Returns
        -------
        LookupNd
        """
        dig, cells = self._build(threads)
        return LookupNd(dig, cells)


def dist_dot(q: "core.Dotprops", t: "core.Dotprops"):
    return list(q.dist_dots(t))


def dist_dot_alpha(q: "core.Dotprops", t: "core.Dotprops"):
    dist, dot, alpha = q.dist_dots(t, alpha=True)
    return [dist, dot * np.sqrt(alpha)]


class LookupDistDotBuilder(LookupNdBuilder):
    """Class for building a 2-dimensional score lookup for NBLAST.

    The scores are

    1. The distances between best-matching points
    2. The dot products of direction vectors around those points,
        optionally scaled by the colinearity `alpha`.

    Parameters
    ----------
    dotprops :          dict or list of Dotprops
                        An indexable sequence of all neurons which will be
                        used as the training set, as Dotprops objects.
    matching_lists :    list of lists of indices into dotprops
                        List of neurons, as indices into `dotprops`, which
                        should be considered matches.
    nonmatching_list :  list of indices into dotprops, optional
                        List of neurons, as indices into `dotprops`,
                        which should not be considered matches.
                        If not given, all `dotprops` will be used
                        (on the assumption that matches are a small subset
                        of possible pairs).
    use_alpha :         bool, optional
                        If true, multiply the dot product by the geometric
                        mean of the matched points' alpha values
                        (i.e. `sqrt(alpha1 * alpha2)`).
    draw_strat :    "batched" | "greedy"
                    Strategy for randomly drawing non-matching pairs.
                    "batched" should be the right choice in most scenarios.
                    "greedy" can be better if your pool of neurons is very
                    small.
    seed :              int, optional
                        Non-matching pairs are drawn at random using this
                        seed, by default {DEFAULT_SEED}.
    """

    def __init__(
        self,
        dotprops: Union[List["core.Dotprops"], Mapping[NeuronKey, "core.Dotprops"]],
        matching_lists: List[List[NeuronKey]],
        nonmatching_list: Optional[List[NeuronKey]] = None,
        use_alpha: bool = False,
        draw_strat: str = "batched",
        seed: int = DEFAULT_SEED,
    ):
        match_fn = dist_dot_alpha if use_alpha else dist_dot
        super().__init__(
            dotprops,
            matching_lists,
            match_fn,
            nonmatching_list,
            draw_strat=draw_strat,
            seed=seed,
        )
        self._ndim = 2

    def build(self, threads=None) -> Lookup2d:
        (dig0, dig1), cells = self._build(threads)
        return Lookup2d(dig0, dig1, cells)


def is_monotonically_increasing(lst):
    for prev_idx, item in enumerate(lst[1:]):
        if item <= lst[prev_idx]:
            return False
    return True


def parse_boundary(item: str):
    explicit_interval = item[0] + item[-1]
    if explicit_interval == "[)":
        right = False
    elif explicit_interval == "(]":
        right = True
    else:
        raise ValueError(
            f"Enclosing characters '{explicit_interval}' do not match a half-open interval"
        )
    return tuple(float(i) for i in item[1:-1].split(",")), right


class LookupAxis(ABC, Generic[T]):
    """Class converting some data into a linear index."""

    @abstractmethod
    def __len__(self) -> int:
        """Number of bins represented by this instance."""
        pass

    @abstractmethod
    def __call__(self, value: Union[T, Sequence[T]]) -> Union[int, Sequence[int]]:
        """Convert some data into a linear index.

        Parameters
        ----------
        value : Union[T, Sequence[T]]
            Value to convert into an index

        Returns
        -------
        Union[int, Sequence[int]]
            If a scalar was given, return a scalar; otherwise, a numpy array of ints.
        """
        pass


class SimpleLookup(LookupAxis[Hashable]):
    """Look up in a list of items and return their index.

    Parameters
    ----------
    items : List[Hashable]
        The item's position in the list is the index which will be returned.

    Raises
    ------
    ValueError
        items are non-unique.
    """

    def __init__(self, items: List[Hashable]):
        self.items = {item: idx for idx, item in enumerate(items)}
        if len(self.items) != len(items):
            raise ValueError("Items are not unique")

    def __len__(self) -> int:
        return len(self.items)

    def __call__(
        self, value: Union[Hashable, Sequence[Hashable]]
    ) -> Union[int, Sequence[int]]:
        if np.isscalar(value):
            return self.items[value]
        else:
            return np.array([self.items[v] for v in value], int)


class Digitizer(LookupAxis[float]):
    """Class converting continuous values into discrete indices.

    Parameters
    ----------
    boundaries : Sequence[float]
        N boundaries specifying N-1 bins.
        Must be monotonically increasing.
    clip : Tuple[bool, bool], optional
        Whether to set the bottom and top boundaries to -infinity and
        infinity respectively, effectively clipping incoming values: by
        default (True, True).
        False means "add a new bin for out-of-range values".
    right : bool, optional
        Whether bins should include their right (rather than left) boundary,
        by default False.
    """

    def __init__(
        self,
        boundaries: Sequence[float],
        clip: Tuple[bool, bool] = (True, True),
        right=False,
    ):
        self.right = right

        boundaries = list(boundaries)
        self._min = -math.inf
        if clip[0]:
            self._min = boundaries[0]
            boundaries[0] = -math.inf
        elif boundaries[0] != -math.inf:
            self._min = -math.inf
            boundaries.insert(0, -math.inf)

        self._max = math.inf
        if clip[1]:
            self._max = boundaries[-1]
            boundaries[-1] = math.inf
        elif boundaries[-1] != math.inf:
            boundaries.append(math.inf)

        if not is_monotonically_increasing(boundaries):
            raise ValueError(
                "Boundaries are not monotonically increasing: " f"{boundaries}"
            )

        self.boundaries = np.asarray(boundaries)

    def __len__(self):
        return len(self.boundaries) - 1

    def __call__(self, value: float):
        # searchsorted is marginally faster than digitize as it skips monotonicity checks
        return (
            np.searchsorted(
                self.boundaries, value, side="left" if self.right else "right"
            )
            - 1
        )

    def to_strings(self, round=None) -> List[str]:
        """Turn boundaries into list of labels.

        Parameters
        ----------
        round :     int, optional
                    Use to round bounds to the Nth decimal.
        """
        if self.right:
            lb = "("
            rb = "]"
        else:
            lb = "["
            rb = ")"

        b = self.boundaries.copy()
        b[0] = self._min
        b[-1] = self._max

        if round:
            b = [np.round(x, round) for x in b]

        return [f"{lb}{lower},{upper}{rb}" for lower, upper in zip(b[:-1], b[1:])]

    @classmethod
    def from_strings(cls, interval_strs: Sequence[str]):
        """Set digitizer boundaries based on a sequence of interval expressions.

        e.g. `["(0, 1]", "(1, 5]", "(5, 10]"]`

        The lowermost and uppermost boundaries are converted to -infinity and
        infinity respectively.

        Parameters
        ----------
        bound_strs : Sequence[str]
            Strings representing intervals, which must abut and have open/closed
            boundaries specified by brackets.

        Returns
        -------
        Digitizer
        """
        bounds: List[float] = []
        last_upper = None
        last_right = None
        for item in interval_strs:
            (lower, upper), right = parse_boundary(item)
            bounds.append(float(lower))

            if last_right is not None:
                if right != last_right:
                    raise ValueError("Inconsistent half-open interval")
            else:
                last_right = right

            if last_upper is not None:
                if lower != last_upper:
                    raise ValueError("Half-open intervals do not abut")

            last_upper = upper

        bounds.append(float(last_upper))
        return cls(bounds, right=last_right)

    @classmethod
    def from_linear(cls, lower: float, upper: float, nbins: int, right=False):
        """Choose digitizer boundaries spaced linearly between two values.

        Input values will be clipped to fit within the given interval.

        Parameters
        ----------
        lower : float
            Lowest value
        upper : float
            Highest value
        nbins : int
            Number of bins
        right : bool, optional
            Whether bins should include their right (rather than left) boundary,
            by default False

        Returns
        -------
        Digitizer
        """
        arr = np.linspace(lower, upper, nbins + 1, endpoint=True)
        return cls(arr, right=right)

    @classmethod
    def from_geom(
        cls, lowest_upper: float, highest_lower: float, nbins: int, right=False
    ):
        """Choose digitizer boundaries in a geometric sequence.

        Additional bins will be added above and below the given values.

        Parameters
        ----------
        lowest_upper : float
            Upper bound of the lowest bin. The lower bound of the lowest bin is
            often 0, which cannot be represented in a nontrivial geometric
            sequence.
        highest_lower : float
            Lower bound of the highest bin.
        nbins : int
            Number of bins
        right : bool, optional
            Whether bins should include their right (rather than left) boundary,
            by default False

        Returns
        -------
        Digitizer
        """
        arr = np.geomspace(lowest_upper, highest_lower, nbins - 1, True)
        return cls(arr, clip=(False, False), right=right)

    @classmethod
    def from_data(
        cls, data: Sequence[float], nbins: int, right=False, method="quantile"
    ):
        """Choose digitizer boundaries to evenly partition the given values.

        Parameters
        ----------
        data : Sequence[float]
            Data which should be partitioned by the resulting digitizer.
        nbins : int
            Number of bins
        right : bool, optional
            Whether bins should include their right (rather than left) boundary,
            by default False
        method : "quantile" | "linear" | "geometric"
            Method to use for partitioning the data space:
             - 'quantile' (default) will partition the data such that each bin
               contains the same number of data points. This is usually the
               method of choice because it is robust against outlier and because
               we are guaranteed to not have empty bin.
             - 'linear' will partition the data into evenly spaced bins.
             - 'geometric' will produce a log scale partition. This will not work
               if data has negative values.

        Returns
        -------
        Digitizer
        """
        assert method in ("quantile", "linear", "geometric")

        if method == "quantile":
            arr = np.quantile(data, np.linspace(0, 1, nbins + 1, True))
        elif method == "linear":
            arr = np.linspace(min(data), max(data), nbins + 1, True)
        elif method == "geometric":
            if min(data) <= 0:
                raise ValueError(
                    "Data must not have values <= 0 for creating "
                    "geometric (logarithmic) bins."
                )
            arr = np.geomspace(min(data), max(data), nbins + 1, True)
        return cls(arr, right=right)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Digitizer):
            return NotImplemented
        return self.right == other.right and np.allclose(
            self.boundaries, other.boundaries
        )


class LookupNd:
    def __init__(self, axes: List[LookupAxis], cells: np.ndarray):
        if [len(b) for b in axes] != list(cells.shape):
            raise ValueError("boundaries and cells have inconsistent bin counts")
        self.axes = axes
        self.cells = cells

    def __call__(self, *args):
        if len(args) != len(self.axes):
            raise TypeError(
                f"Lookup takes {len(self.axes)} arguments but {len(args)} were given"
            )

        idxs = tuple(d(arg) for d, arg in zip(self.axes, args))
        out = self.cells[idxs]
        return out


class Lookup2d(LookupNd):
    """Convenience class inheriting from LookupNd for the common 2D float case.
    Provides IO with pandas DataFrames.
    """

    def __init__(self, axis0: Digitizer, axis1: Digitizer, cells: np.ndarray):
        """2D lookup table for convert NBLAST matches to scores.

        Commonly read from a `pandas.DataFrame`
        or trained on data using a `LookupDistDotBuilder`.

        Parameters
        ----------
        axis0 : Digitizer
            How to convert continuous values into an index for the first axis.
        axis1 : Digitizer
            How to convert continuous values into an index for the second axis.
        cells : np.ndarray
            Values to look up in the table.
        """
        super().__init__([axis0, axis1], cells)

    def to_dataframe(self) -> pd.DataFrame:
        """Convert the lookup table into a `pandas.DataFrame`.

        From there, it can be shared, saved, and so on.

        The index and column labels describe the intervals represented by that axis.

        Returns
        -------
        pd.DataFrame
        """
        return pd.DataFrame(
            self.cells,
            self.axes[0].to_strings(),
            self.axes[1].to_strings(),
        )

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame):
        """Parse score matrix from a dataframe with string index and column labels.

        Expects the index and column labels to specify an interval
        like `f"[{{lower}},{{upper}})"`.
        Will replace the lowermost and uppermost bound with -inf and inf
        if they are not already.
        """
        return cls(
            Digitizer.from_strings(df.index),
            Digitizer.from_strings(df.columns),
            df.to_numpy(),
        )


@lru_cache(maxsize=None)
def _smat_fcwb(alpha=False):
    # cached private function defers construction
    # until needed (speeding startup),
    # but avoids repeated reads (speeding later uses)
    fname = ("smat_fcwb.csv", "smat_alpha_fcwb.csv")[alpha]
    fpath = smat_path / fname

    return Lookup2d.from_dataframe(pd.read_csv(fpath, index_col=0))


def smat_fcwb(alpha=False):
    # deepcopied so that mutations do not propagate to cache
    return deepcopy(_smat_fcwb(alpha))


def check_score_fn(fn: Callable, nargs=2, scalar=True, array=True):
    """Checks functionally that the callable can be used as a score function.

    Parameters
    ----------
    nargs : optional int, default 2
        How many positional arguments the score function should have.
    scalar : optional bool, default True
        Check that the function can be used on `nargs` scalars.
    array : optional bool, default True
        Check that the function can be used on `nargs` 1D `numpy.ndarray`s.

    Raises
    ------
    ValueError
        If the score function is not appropriate.
    """
    if scalar:
        scalars = [0.5] * nargs
        if not isinstance(fn(*scalars), float):
            raise ValueError("smat does not take 2 floats and return a float")

    if array:
        test_arr = np.array([0.5] * 3)
        arrs = [test_arr] * nargs
        try:
            out = fn(*arrs)
        except Exception as e:
            raise ValueError(f"Failed to use smat with numpy arrays: {e}")

        if out.shape != test_arr.shape:
            raise ValueError(
                f"smat produced inconsistent shape: input {test_arr.shape}; output {out.shape}"
            )


SCORE_FN_DESCR = """
NBLAST score functions take 2 floats or N-length numpy arrays of floats
(for matched dotprop points/tangents, distance and dot product;
the latter possibly scaled by the geometric mean of the alpha colinearity values)
and returns a float or N-length numpy array of floats.
""".strip().replace("\n", " ")


def parse_score_fn(smat, alpha=False):
    f"""Interpret `smat` as a score function.
    Primarily for backwards compatibility.
    {SCORE_FN_DESCR}
    Parameters
    ----------
    smat : None | "auto" | str | os.PathLike | pandas.DataFrame | Callable[[float, float], float]
        If `None`, use `operator.mul`.
        If `"auto"`, use `navis.nbl.smat.smat_fcwb(alpha)`.
        If a dataframe, use `navis.nbl.smat.Lookup2d.from_dataframe(smat)`.
        If another string or path-like, load from CSV in a dataframe and uses as above.
        Also checks the signature of the callable.
        Raises an error, probably a ValueError, if it can't be interpreted.
    alpha : optional bool, default False
        If `smat` is `"auto"`, choose whether to use the FCWB matrices
        with or without alpha.
    Returns
    -------
    Callable
    Raises
    ------
    ValueError
        If score function cannot be interpreted.
    """
    if smat is None:
        smat = operator.mul
    elif smat == "auto":
        smat = smat_fcwb(alpha)

    if isinstance(smat, (str, os.PathLike)):
        smat = pd.read_csv(smat, index_col=0)

    if isinstance(smat, pd.DataFrame):
        smat = Lookup2d.from_dataframe(smat)

    if not callable(smat):
        raise ValueError(
            "smat should be a callable, a path, a pandas.DataFrame, or 'auto'"
        )

    check_score_fn(smat)

    return smat



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import numpy as np
import pandas as pd

from scipy import ndimage
from typing import Optional, Union, List

from .. import config, graph, core, utils, meshes
from .utils import sample_points_uniform

# Set up logging
logger = config.get_logger(__name__)

__all__ = ["downsample_neuron"]


@utils.map_neuronlist(desc="Downsampling", allow_parallel=True)
def downsample_neuron(
    x: "core.NeuronObject",
    downsampling_factor: Union[int, float],
    inplace: bool = False,
    preserve_nodes: Optional[List[int]] = None,
) -> Optional["core.NeuronObject"]:
    """Downsample neuron(s) by a given factor.

    For skeletons: preserves root, leafs, branchpoints by default. Preservation
    of nodes with synapses can be toggled - see `preserve_nodes` parameter.
    Use `downsampling_factor=float('inf')` to get a skeleton consisting only
    of root, branch and end points.

    Parameters
    ----------
    x :                     single neuron | NeuronList
                            Neuron(s) to downsample. Note that for MeshNeurons
                            we use the first available backend.
    downsampling_factor :   int | float('inf')
                            Factor by which downsample. For TreeNeuron, Dotprops
                            and MeshNeurons this reduces the node, point
                            and face count, respectively. For VoxelNeurons it
                            reduces the dimensions by given factor.
    preserve_nodes :        str | list, optional
                            Can be either list of node IDs to exclude from
                            downsampling or a string to a DataFrame attached
                            to the neuron (e.g. "connectors"). DataFrame must
                            have `node_id` column. Only relevant for
                            TreeNeurons.
    inplace :               bool, optional
                            If True, will modify original neuron. If False, we
                            will operate and return o a copy.

    Returns
    -------
    TreeNeuron/Dotprops/VoxelNeurons/NeuronList
                            Same datatype as input.

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> n_ds = navis.downsample_neuron(n,
    ...                                downsampling_factor=5,
    ...                                inplace=False)
    >>> n.n_nodes > n_ds.n_nodes
    True

    See Also
    --------
    [`navis.resample_skeleton`][]
                             This function resamples a neuron to given
                             resolution. This will change node IDs!
    [`navis.simplify_mesh`][]
                             This is the function used for `MeshNeurons`. Use
                             directly for more control of the simplification.

    """
    if downsampling_factor <= 1:
        raise ValueError('Downsampling factor must be greater than 1.')

    if not inplace:
        x = x.copy()

    if isinstance(x, core.TreeNeuron):
        _ = _downsample_treeneuron(x,
                                   downsampling_factor=downsampling_factor,
                                   preserve_nodes=preserve_nodes)
    elif isinstance(x, core.Dotprops):
        _ = _downsample_dotprops(x,
                                 downsampling_factor=downsampling_factor)
    elif isinstance(x, core.VoxelNeuron):
        _ = _downsample_voxels(x,
                               downsampling_factor=downsampling_factor)
    elif isinstance(x, core.MeshNeuron):
        _ = meshes.simplify_mesh(x,
                                 F=1/downsampling_factor,
                                 inplace=True)
    else:
        raise TypeError(f'Unable to downsample data of type "{type(x)}"')

    return x


def _downsample_voxels(x, downsampling_factor, order=1):
    """Downsample voxels."""
    assert isinstance(x, core.VoxelNeuron)

    zoom_factor = 1 / downsampling_factor

    # order=1 means linear interpolation
    x._data = ndimage.zoom(x.grid, zoom_factor, order=order)

    # We have to change the units here too
    x.units *= downsampling_factor


def _downsample_dotprops(x, downsampling_factor, method="simple"):
    """Downsample Dotprops."""
    assert isinstance(x, core.Dotprops)

    # Can't downsample if no points
    if isinstance(x._points, type(None)):
        return

    # If not enough points
    if x._points.shape[0] <= downsampling_factor:
        return

    # Generate a mask
    if method == "simple":
        mask = np.arange(0, x._points.shape[0], int(downsampling_factor))
    elif method == "uniform":
        mask = sample_points_uniform(
            x._points, x._points.shape[0] // downsampling_factor, output="mask"
        )
    else:
        raise ValueError(f"Unknown (down-)sampling method: {method}")

    # Mask vectors
    # This will also trigger re-calculation which is necessary for two reasons:
    # 1. Vectors will change dramatically if they have to be recalculated from
    #    the downsampled dotprops.
    # 2. There might not be enough points left after downsampling given the
    #    original k.
    if isinstance(x._vect, type(None)) and x.k:
        x.recalculate_tangents(k=x.k, inplace=True)
    x._vect = x._vect[mask]

    # Mask alphas if exists
    if not isinstance(x._alpha, type(None)):
        x._alpha = x._alpha[mask]

    # Finally mask points
    x._points = x._points[mask]


def _downsample_treeneuron(x, downsampling_factor, preserve_nodes):
    """Downsample TreeNeuron."""
    assert isinstance(x, core.TreeNeuron)

    if not isinstance(preserve_nodes, type(None)):
        if isinstance(preserve_nodes, str):
            table = getattr(x, preserve_nodes)
            if not isinstance(table, pd.DataFrame):
                raise TypeError(f'Expected "{preserve_nodes}" to be a '
                                f'DataFrame - got {type(table)}')
            if 'node_id' not in table.columns:
                raise IndexError(f'DataFrame {preserve_nodes} has no "node_id"'
                                 ' column.')

            preserve_nodes = table['node_id'].values

        if not isinstance(preserve_nodes, (list, set, np.ndarray)):
            raise TypeError('Expected "preserve_nodes" to be list-like, got '
                            f'"{type(preserve_nodes)}"')

    if x.nodes.shape[0] <= 1:
        logger.warning(f'Neuron {x.id} has no nodes. Skipping.')
        return

    list_of_parents = {n: p for n, p in zip(x.nodes.node_id.values,
                                            x.nodes.parent_id.values)}
    list_of_parents[-1] = -1  # type: ignore  # doesn't know that node_id is int

    if 'type' not in x.nodes:
        graph.classify_nodes(x)

    selection = x.nodes.type != 'slab'

    if utils.is_iterable(preserve_nodes):
        selection = selection | x.nodes.node_id.isin(preserve_nodes)  # type: ignore

    fix_points = x.nodes[selection].node_id.values

    # Add soma node(s)
    if not isinstance(x.soma, type(None)):
        soma = utils.make_iterable(x.soma)
        for s in soma:
            if s not in fix_points:
                fix_points = np.append(fix_points, s)

    # Walk from all fix points to the root - jump N nodes on the way
    new_parents = {}

    for en in fix_points:
        this_node = en

        while True:
            stop = False
            new_p = list_of_parents[this_node]
            if new_p >= 0:
                i = 0
                while i < downsampling_factor:
                    if new_p in fix_points or new_p < 0:
                        new_parents[this_node] = new_p
                        stop = True
                        break
                    new_p = list_of_parents[new_p]
                    i += 1

                if stop is True:
                    break
                else:
                    new_parents[this_node] = new_p
                    this_node = new_p
            else:
                new_parents[this_node] = -1  # type: ignore
                break

    # Subset to kept nodes
    new_nodes = x.nodes[x.nodes.node_id.isin(list(new_parents.keys()))].copy()

    # Assign new parent IDs
    new_nodes['parent_id'] = new_nodes.node_id.map(new_parents).astype(int)

    logger.debug(f'Nodes before/after: {len(x.nodes)}/{len(new_nodes)}')

    x.nodes = new_nodes

    # This is essential -> otherwise e.g. graph.neuron2graph will fail
    x.nodes.reset_index(inplace=True, drop=True)

    x._clear_temp_attr()



"""Configuration for submodule doctests.

Should be importable (but not useful)
without development dependencies.
"""
try:
    import vispy
except ModuleNotFoundError:
    collect_ignore_glob = ["*.py"]



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2017 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import uuid
import functools

from .. import config, core, io

# Set up logging
logger = config.get_logger(__name__)


def patch_cloudvolume():
    """Monkey patch cloud-volume to return navis neurons.

    This function must be run before initializing the `CloudVolume`! Adds new
    methods/parameters to `CloudVolume.mesh.get` and `CloudVolume.skeleton.get`.
    See examples for details.

    Examples
    --------
    >>> import navis
    >>> import cloudvolume as cv
    >>> # Monkey patch cloudvolume
    >>> navis.patch_cloudvolume()
    >>> # Connect to the Google segmentation of FAFB
    >>> vol = cv.CloudVolume('precomputed://gs://fafb-ffn1-20200412/segmentation',
    ...                       use_https=True,
    ...                       progress=False)
    >>> ids = [2137190164, 2268989790]
    >>> # Fetch as navis neuron using newly added method or ...
    >>> nl = vol.mesh.get_navis(ids, lod=3)
    >>> # ... alternatively use `as_navis` keyword argument in original method
    >>> nl = vol.mesh.get(ids, lod=3, as_navis=True)
    >>> type(nl)
    <class 'navis.core.neuronlist.NeuronList'>
    >>> # The same works for skeletons
    >>> skels = vol.skeleton.get_navis(ids)

    """
    global cv
    try:
        import cloudvolume as cv
    except ModuleNotFoundError:
        cv = None

    # If CV not installed do nothing
    if not cv:
        logger.info("cloud-volume appears to not be installed?")
        return

    for ds in [
        cv.datasource.graphene.mesh.sharded.GrapheneShardedMeshSource,
        cv.datasource.graphene.mesh.unsharded.GrapheneUnshardedMeshSource,
        cv.datasource.precomputed.mesh.unsharded.UnshardedLegacyPrecomputedMeshSource,
        cv.datasource.precomputed.mesh.multilod.UnshardedMultiLevelPrecomputedMeshSource,
        cv.datasource.precomputed.mesh.multilod.ShardedMultiLevelPrecomputedMeshSource,
        cv.datasource.precomputed.skeleton.sharded.ShardedPrecomputedSkeletonSource,
        cv.datasource.precomputed.skeleton.unsharded.UnshardedPrecomputedSkeletonSource,
    ]:
        ds.get_navis = return_navis(ds.get, only_on_kwarg=False)
        ds.get = return_navis(ds.get, only_on_kwarg=True)

    logger.info("cloud-volume successfully patched!")


def return_navis(func, only_on_kwarg=False):
    """Wrap cloud-volume mesh and skeleton sources.

    Parameters
    ----------
    func :          callable
                    Function/method to wrap.
    only_on_kwarg : bool
                    If True, will look for a `as_navis=True` (default=False)
                    keyword argument to determine if results should be converted
                    to navis neurons. If 'process' is set to False, the neuron
                    will not be processed by TriMesh (remove nan, duplicate vertices,etc)

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        ret_navis = kwargs.pop("as_navis", False)
        process = kwargs.pop("process", False)
        res = func(*args, **kwargs)

        if not only_on_kwarg or ret_navis:
            neurons = []
            if isinstance(res, (list, tuple)):
                res = {getattr(n, "id", uuid.uuid4()): n for n in res}
            if isinstance(res, (cv.Mesh, cv.Skeleton)):
                res = {getattr(res, "id", uuid.uuid4()): res}

            for k, v in res.items():
                if isinstance(v, cv.Mesh):
                    n = core.MeshNeuron(v, id=k, units="nm", process=process)
                    neurons.append(n)
                elif isinstance(v, cv.Skeleton):
                    swc_str = v.to_swc()
                    n = io.read_swc(swc_str)
                    n.id = k
                    n.units = "nm"
                    neurons.append(n)
                else:
                    logger.warning(
                        f"Skipped {k}: Unable to convert {type(v)} to " "navis Neuron."
                    )

            return core.NeuronList(neurons)
        return res

    return wrapper



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import numpy as np
import networkx as nx

try:
    from pykdtree.kdtree import KDTree
except ModuleNotFoundError:
    from scipy.spatial import cKDTree as KDTree


def sample_points_uniform(points, size, output="points"):
    """Draw uniform sample from point cloud.

    This functions works by iteratively removing the point with the smallest
    distance to its nearest neighbor until the desired number of points is
    reached.

    Parameters
    ----------
    points :    (N, 3 ) array
                Point cloud to sample from.
    size :      int
                Number of samples to draw.
    output :    "points" | "indices" | "mask", optional
                If "points", returns the sampled points. If "indices", returns
                the indices of the sampled points. If "mask", returns a boolean
                mask of the sampled points.

    Returns
    -------
    See `output` parameter.

    """
    points = np.asarray(points)

    assert isinstance(points, np.ndarray) and points.ndim == 2 and points.shape[1] == 3
    assert output in ("points", "indices", "mask")
    assert (size > 0) and (size <= len(points))

    # Start with all points in the mask
    mask = np.ones(len(points), dtype=bool)

    # Generate a tree
    tree = KDTree(points)

    p_ind = np.arange(len(points))

    while mask.sum() > size:
        # Find the point with the largest distance to its nearest neighbor
        d, ind = tree.query(points[mask], k=2, mask=~mask)
        d, ind = d[:, 1], ind[:, 1]

        # Find pairs of nodes that are close to each other
        is_close = d == d.min()
        pairs = np.stack((p_ind[mask][is_close], p_ind[ind][is_close]), axis=1)

        # At this point we will have pairs show up multiple times - (a, b) and (b, a)
        pairs = np.unique(np.sort(pairs, axis=1), axis=0)

        # Imagine we have two candidate pairs for removal: (a, b) and (b, c)
        # In that case we can remove (a and c) or (b) but not (a, b) or (b, c)
        # because that might leave a hole in the point cloud
        G = nx.Graph()
        G.add_edges_from(pairs)

        to_remove = []
        for cc in nx.connected_components(G):
            # If these are two nodes, it doesn't matter which one we drop
            if len(cc) <= 2:
                to_remove.append(cc.pop())
                continue
            # If we have three or more nodes, we will simply remove the one
            # with the highest degree
            to_remove.append(sorted(cc, key=lambda x: G.degree(x))[-1])

        # Number of nodes we still need to remove
        n_remove = mask.sum() - size

        if n_remove >= len(to_remove):
            mask[to_remove] = False
        else:
            mask[to_remove[:n_remove]] = False

    if output == "mask":
        return mask
    elif output == "indices":
        return p_ind[mask]
    elif output == "points":
        return points[mask].copy()







